# a note about this file: i've appended a "; raise" at the end of nearly every custom raise_error_at_token call, 
# because pyright (my nvim LSP) doesn't recognize the code terminating at the raise_error_at_token, so I add this 
# to make sure it recognizes that and doesn't yell at me because I don't like being yelled at

# thanks to Indently, I have now discovered the NoReturn type, and pyright shuts up :))))))))))))))))

from __future__ import annotations
import os
import re
import locale
import random
import pickle
import requests
from time import sleep
from pathlib import Path
from copy import deepcopy
from threading import Thread
from difflib import SequenceMatcher
from typing import Literal, Optional, TypeAlias, Union 

KEY_MOUSE_IMPORTED = True
try: 
    from pynput import keyboard, mouse
except ImportError:
    KEY_MOUSE_IMPORTED = False

GITHUB_IMPORTED = True 
try: 
    import github 
except ImportError:
    GITHUB_IMPORTED = False

from dreamberd.base import NonFormattedError, OperatorType, Token, TokenType, debug_print, debug_print_no_token, raise_error_at_line, raise_error_at_token
from dreamberd.builtin import FLOAT_TO_INT_PREC, BuiltinFunction, DreamberdBoolean, DreamberdFunction, DreamberdIndexable, DreamberdKeyword, DreamberdList, DreamberdMap, DreamberdMutable, DreamberdNamespaceable, DreamberdNumber, DreamberdObject, DreamberdPromise, DreamberdSpecialBlankValue, DreamberdString, DreamberdUndefined, Name, Variable, Value, VariableLifetime, db_not, db_to_boolean, db_to_number, db_to_string, is_int
from dreamberd.processor.lexer import tokenize as db_tokenize
from dreamberd.processor.expression_tree import ExpressionTreeNode, FunctionNode, ListNode, SingleOperatorNode, ValueNode, IndexNode, ExpressionNode, build_expression_tree, get_expr_first_token
from dreamberd.processor.syntax_tree import AfterStatement, ClassDeclaration, CodeStatement, CodeStatementKeywordable, Conditional, DeleteStatement, ExportStatement, ExpressionStatement, FunctionDefinition, ImportStatement, ReturnStatement, ReverseStatement, VariableAssignment, VariableDeclaration, WhenStatement

# several "ratios" used in the approx equal function
NUM_EQUALITY_RATIO = 0.1  # a-b / b 
STRING_EQUALITY_RATIO = 0.7  # min ratio to be considered equal
LIST_EQUALITY_RATIO = 0.7  # min ratio of all the elements of a list to be equal for the lists to be equal
MAP_EQUALITY_RATIO = 0.6  # lower thresh cause i feel like it
FUNCTION_EQUALITY_RATIO = 0.6  # yeah 
OBJECT_EQUALITY_RATIO = 0.6 

# thing used in the .dreamberd_runtime file
DB_RUNTIME_PATH = ".dreamberd_runtime"
INF_VAR_PATH = ".inf_vars"
INF_VAR_VALUES_PATH = ".inf_vars_values"
DB_VAR_TO_VALUE_SEP = ";;;"  # i'm feeling fancy

# :D 
Namespace: TypeAlias = dict[str, Union[Variable, Name]]
CodeStatementWithExpression: TypeAlias = Union[ReturnStatement, Conditional, ExpressionStatement, WhenStatement,
                                               VariableAssignment, AfterStatement, VariableDeclaration]
AsyncStatements: TypeAlias = list[tuple[list[tuple[CodeStatement, ...]], list[Namespace], int, Union[Literal[1], Literal[-1]]]]
NameWatchers: TypeAlias = dict[tuple[str, int], tuple[CodeStatementWithExpression, set[tuple[str, int]], list[Namespace], Optional[DreamberdPromise]]]
WhenStatementWatchers: TypeAlias = list[dict[Union[str, int], list[tuple[ExpressionTreeNode, list[tuple[CodeStatement, ...]]]]]]  # bro there are six square brackets...

def get_built_expression(expr: Union[list[Token], ExpressionTreeNode]) -> ExpressionTreeNode:
    return expr if isinstance(expr, ExpressionTreeNode) else build_expression_tree(filename, expr, code)

def get_modified_next_name(name: str, ns: int) -> str:
    return f"{name}_{ns}__next"

def get_modified_prev_name(name: str) -> str:
    return f"{name.replace('.', '__')}__prev"

# i believe this function is exclusively called from the evaluate_expression function
def evaluate_normal_function(expr: FunctionNode, func: Union[DreamberdFunction, BuiltinFunction], namespaces: list[Namespace], args: list[Value], when_statement_watchers: WhenStatementWatchers) -> Value:

    # check to evaluate builtin
    if isinstance(func, BuiltinFunction):
        if func.arg_count > len(args):
            raise_error_at_token(filename, code, f"Expected more arguments for function call with {func.arg_count} argument{'s' if func.arg_count != 1 else ''}.", expr.name)
        max_arg_count = func.arg_count if func.arg_count >= 0 else len(args)
        return func.function(*args[:max_arg_count]) or DreamberdUndefined()
    
    # check length is proper, adjust namespace, and run this code
    if len(func.args) > len(args):
        raise_error_at_token(filename, code, f"Expected more arguments for function call with {len(func.args)} argument{'s' if len(func.args) != 1 else ''}.", expr.name)
    new_namespace: Namespace = {name: Name(name, arg) for name, arg in zip(func.args, args)}
    return interpret_code_statements(func.code, namespaces + [new_namespace], [], when_statement_watchers + [{}]) or DreamberdUndefined()

def register_async_function(expr: FunctionNode, func: DreamberdFunction, namespaces: list[Namespace], args: list[Value], async_statements: AsyncStatements) -> None:
    """ Adds a job to the async statements queue, which is accessed in the interpret_code_statements function. """
    if len(func.args) > len(args):
        raise_error_at_token(filename, code, f"Expected more arguments for function call with {len(func.args)} argument{'s' if len(func.args) != 1 else ''}.", expr.name)
    function_namespaces = namespaces + [{name: Name(name, arg) for name, arg in zip(func.args, args)}]
    async_statements.append((func.code, function_namespaces, 0, 1))

def get_code_from_when_statement_watchers(name_or_id: Union[str, int], when_statement_watchers: WhenStatementWatchers) -> list[tuple[ExpressionTreeNode, list[tuple[CodeStatement, ...]]]]:
    vals = []
    for watcher_dict in when_statement_watchers:
        if val := watcher_dict.get(name_or_id):
            vals += val
    return vals

def remove_from_when_statement_watchers(name_or_id: Union[str, int], watcher: tuple[ExpressionTreeNode, list[tuple[CodeStatement, ...]]], when_statement_watchers: WhenStatementWatchers) -> None:
    for watcher_dict in when_statement_watchers:
        if vals := watcher_dict.get(name_or_id):
            remove = None
            for i, v in enumerate(vals):
                if v == watcher:
                    remove = i
            if remove is not None: 
                del vals[remove]

def remove_from_all_when_statement_watchers(name_or_id: Union[str, int], when_statement_watchers: WhenStatementWatchers) -> None: 
    for watcher_dict in when_statement_watchers:
        if name_or_id in watcher_dict:
            del watcher_dict[name_or_id]

def load_global_dreamberd_variables(namespaces: list[Namespace]) -> None:

    dir_path = Path().home()/DB_RUNTIME_PATH
    inf_values_path = dir_path/INF_VAR_VALUES_PATH
    inf_var_list = dir_path/INF_VAR_PATH
    if not dir_path.is_dir(): return 
    if not inf_values_path.is_dir(): return 
    if not inf_var_list.is_file(): return 

    with open(inf_var_list, 'r') as f:
        for line in f.readlines():
            if not line.strip():
                continue
        
            name, identity, can_be_reset, can_edit_value, confidence = line.split(DB_VAR_TO_VALUE_SEP)
            can_be_reset = eval(can_be_reset) if can_be_reset in ["True", "False"] else True # safe code !!!!!!!!!!!!
            can_edit_value = eval(can_edit_value) if can_edit_value in ["True", "False"] else True

            with open(dir_path/INF_VAR_VALUES_PATH/identity, "rb") as data_f:
                value = pickle.load(data_f)
            namespaces[-1][name] = Variable(name, [VariableLifetime(value, 100000000000, int(confidence), can_be_reset, can_edit_value)], [])

def load_public_global_variables(namespaces: list[Namespace]) -> None:
    repo_url = "https://raw.githubusercontent.com/vivaansinghvi07/dreamberd-interpreter-globals/main"
    for line in requests.get(f"{repo_url}/public_globals.txt").text.split("\n"):
        if not line.strip(): continue
        name, address, confidence = line.split(DB_VAR_TO_VALUE_SEP)
        can_be_reset = can_edit_value = False  # these were const 

        encoded_value = requests.get(f"{repo_url}/global_objects/{address}").text
        byte_list = [int(encoded_value[i:i+4]) for i in range(0, len(encoded_value), 4)]
        value = pickle.loads(bytearray(byte_list))
        namespaces[-1][name] = Variable(name, [VariableLifetime(value, 100000000000, int(confidence), can_be_reset, can_edit_value)], [])

def open_global_variable_issue(name: str, value: Value, confidence: int):
    if not GITHUB_IMPORTED:
        raise_error_at_line(filename, code, current_line, "Cannot create a public global variable without a the GitHub API imported.")
    try:
        access_token = os.environ["GITHUB_ACCESS_TOKEN"]
    except KeyError:
        raise_error_at_line(filename, code, current_line, "To declare public globals, you must set the GITHUB_ACCESS_TOKEN to a personal access token.")

    # transform the variable into a value string
    value_bytes = list(pickle.dumps(value))
    issue_body = "".join([str(b).rjust(4, '0') for b in value_bytes])

    # post the variable as an issue to the main github repo
    with github.Github(auth=github.Auth.Token(access_token)) as g:   # type: ignore 
        repo = g.get_repo("vivaansinghvi07/dreamberd-interpreter-globals")
        repo.create_issue(f"Create Public Global: {name}{DB_VAR_TO_VALUE_SEP}{confidence}", issue_body)

def declare_new_variable(statement: VariableDeclaration, value: Value, namespaces: list[Namespace], async_statements: AsyncStatements, when_statement_watchers: WhenStatementWatchers):

    name, lifetime, confidence, debug, modifiers = statement.name.value, statement.lifetime, statement.confidence, statement.debug, statement.modifiers 
    name_token = statement.name  # for error handling purposes
    is_global = len(modifiers) == 3
    can_be_reset = isinstance(v := get_value_from_namespaces(modifiers[-2], namespaces), DreamberdKeyword) and v.value == "var"
    can_edit_value = isinstance(v := get_value_from_namespaces(modifiers[-1], namespaces), DreamberdKeyword) and v.value == "var"

    if '.' in name:
        raise_error_at_token(filename, code, "Cannot declare a variable with periods in the name.", name_token)

    is_lifetime_temporal = lifetime is not None and not lifetime[-1].isdigit()
    variable_duration = 100000000000 if is_lifetime_temporal or lifetime is None else int(lifetime)
    target_lifetime = VariableLifetime(value, variable_duration, confidence, can_be_reset, can_edit_value)

    if v := namespaces[-1].get(name): 
        if isinstance(v, Variable):   # check for another declaration?
            target_var = v
            for i in range(len(v.lifetimes) + 1):
                if i == len(v.lifetimes) or v.lifetimes[i].confidence == confidence:
                    if i == 0:
                        v.prev_values.append(v.value)
                    v.lifetimes[i:i] = [target_lifetime]
        else:
            target_var = Variable(name, [target_lifetime], [v.value])
            namespaces[-1][name] = target_var
    else:  # for loop finished unbroken, no matches found
        target_var = Variable(name, [target_lifetime], [])
        namespaces[-1][name] = target_var

    match debug:
        case 0: pass 
        case 1: 
            debug_print(filename, code, f"Setting {statement.name.value} to {db_to_string(value).value}", statement.name)
        case 2: 
            expr = get_built_expression(statement.expression)
            debug_print(filename, code, f"Setting {' '.join([mod.value for mod in statement.modifiers])} variable \"{statement.name.value}\" to {db_to_string(value).value} with a lifetime of {lifetime}.", statement.name)
        case 3: 
            expr = get_built_expression(statement.expression)
            names = gather_names_or_values(expr)
            debug_print(filename, code, f"Setting {' '.join([mod.value for mod in statement.modifiers])} variable \"{statement.name.value}\" to {db_to_string(value).value} with a lifetime of {lifetime}.\nThe value of each name in the expression is the following: \n{chr(10).join([f'  {name}: {db_to_string(get_value_from_namespaces(name_token, namespaces)).value}' for name in names])}", statement.name)
        case _: 
            expr = get_built_expression(statement.expression)
            names = gather_names_or_values(expr)
            debug_print(filename, code, f"Setting {' '.join([mod.value for mod in statement.modifiers])} variable \"{statement.name.value}\" to {db_to_string(value).value} with a lifetime of {lifetime}.\nThe value of each name in the exprehttps://accounts.spotify.com/en/login?continue=https%3A%2F%2Fopen.spotify.com%2Fplaylist%2F0jqXRIN34QAfwEiB72S4H7%3Fsi%3D6f53123059f741ec%26pt%3D09c9c4118ece8f2d0fad56d9f2a4345ession is the following: \n{chr(10).join([f'  {name}: {db_to_string(get_value_from_namespaces(name_token, namespaces)).value}' for name in names])}\nThe expression used to get this value is: \n{expr.to_string()}", statement.name)

    # check if there is a watcher for this name
    watchers_key = (name, id(namespaces[-1]))
    if watcher := name_watchers.get(watchers_key):
        st, stored_nexts, watcher_ns, promise = watcher
        mod_name = get_modified_next_name(*watchers_key)
        watcher_ns[-1][mod_name] = Name(mod_name, value)  # add the value to the uppermost namespace
        stored_nexts.remove(watchers_key)                   # remove the name from the set containing remaining names
        if not stored_nexts:  # not waiting on anybody else, execute the code
            interpret_name_watching_statement(st, watcher_ns, promise, async_statements, when_statement_watchers)
        del name_watchers[watchers_key]             # stop watching this name

    # check if this name appears in a when statement of the appropriate scope  --  it would have to be watching the name
    if when_watchers := get_code_from_when_statement_watchers(name, when_statement_watchers):
        for when_watcher in when_watchers:  # i just wanna be done with this :(
            condition, inside_statements = when_watcher
            condition_val = evaluate_expression(condition, namespaces, async_statements, when_statement_watchers)
            if isinstance(value, DreamberdMutable):
                when_statement_watchers[-1][id(value)].append(when_watcher)  ##### remember : this is tuple so it is immutable and copied !!!!!!!!!!!!!!!!!!!!!!  # wait nvm i suick at pytghon
            if isinstance(target_var.prev_values[-1], DreamberdMutable):   # if prev value was being observed under this statement, remove it  ??
                remove_from_when_statement_watchers(id(target_var.prev_values[-1]), when_watcher, when_statement_watchers)
            when_statement_watchers[-1][id(target_var)].append(when_watcher)  # put this where the new variable is
            execute_conditional(condition_val, inside_statements, namespaces, when_statement_watchers)
        remove_from_all_when_statement_watchers(name, when_statement_watchers)  # that name is now set to a variable, discard it from the when statement  --  it is now a var not a string

    if is_global:
        open_global_variable_issue(name, value, confidence)

    # if we're dealing with seconds just sleep in another thread and remove the variable lifetime
    if lifetime == "Infinity":
        # if len(namespaces) == 1: continue  # only save global vars if they are in the global scope

        # define and initialize the directories
        dir_path = Path().home()/DB_RUNTIME_PATH
        inf_values_path = dir_path/INF_VAR_VALUES_PATH
        if not dir_path.is_dir(): dir_path.mkdir()
        if not inf_values_path.is_dir(): inf_values_path.mkdir()

        generated_addr = random.randint(1, 100000000000)  # hopefully never repeat, if it does, oh well :)
        with open(dir_path/INF_VAR_PATH, 'a') as f:
            SEP = DB_VAR_TO_VALUE_SEP
            f.write(f"{name}{SEP}{generated_addr}{SEP}{can_be_reset}{SEP}{can_edit_value}{SEP}{confidence}\n")
        with open(dir_path/INF_VAR_VALUES_PATH/str(generated_addr), "wb") as f:
            pickle.dump(value, f)
            
    elif is_lifetime_temporal:
        def remove_lifetime(lifetime: str, target_var: Variable, target_lifetime: VariableLifetime, error_line: int):
            if lifetime[-1] not in ['s', 'm'] or all(c.isdigit() for c in lifetime[:-1]):
                raise_error_at_line(filename, code, error_line, "Invalid time unit for variable lifetime.")
            sleep(int(lifetime[:-1]) if lifetime[-1] == 's' else int(lifetime[:-1] * 60))
            for i, lt in reversed([*enumerate(target_var.lifetimes)]):
                if lt is target_lifetime:
                    del target_var.lifetimes[i]
        Thread(target=remove_lifetime, args=(lifetime, target_var, target_lifetime)).start()

def assign_variable(statement: VariableAssignment, indexes: list[Value], new_value: Value, namespaces: list[Namespace], async_statements: AsyncStatements, when_statement_watchers: WhenStatementWatchers):
    name, confidence, debug = statement.name.value, statement.confidence, statement.debug
    name_token = statement.name
        
    var, ns = get_name_and_namespace_from_namespaces(name, namespaces)
    if var is None:
        raise_error_at_token(filename, code, "Attempted to set a name that is undefined.", name_token)
 
    match debug:
        case 0: pass 
        case 1: 
            debug_print(filename, code, f"Setting {statement.name.value}{''.join([f'[{db_to_string(val).value}]' for val in indexes])} to {db_to_string(new_value).value}", statement.name)
        case 2:
            expr = get_built_expression(statement.expression)
            names = gather_names_or_values(expr)
            debug_print(filename, code, f"Setting {statement.name.value}{''.join([f'[{db_to_string(val).value}]' for val in indexes])} to {db_to_string(new_value).value}\nThe value of each name in the expression is the following: \n{chr(10).join([f'  {name}: {db_to_string(get_value_from_namespaces(name, namespaces)).value}' for name in names])}", statement.name)
        case 3: 
            expr = get_built_expression(statement.expression)
            names = gather_names_or_values(expr)
            debug_print(filename, code, f"Setting {statement.name.value}{''.join([f'[{db_to_string(val).value}]' for val in indexes])} to {db_to_string(new_value).value}\nThe value of each name in the expression is the following: \n{chr(10).join([f'  {name}: {db_to_string(get_value_from_namespaces(name, namespaces)).value}' for name in names])}\nThe expression used to get this value is: \n{expr.to_string()}", statement.name)
        case _:
            expr = get_built_expression(statement.expression)
            index_exprs = [get_built_expression(ex) for ex in statement.indexes]
            names = gather_names_or_values(expr)
            for ex in index_exprs:
                names |= gather_names_or_values(ex)
            debug_print(filename, code, f"Setting {statement.name.value}{''.join([f'[{db_to_string(val).value}]' for val in indexes])} to {db_to_string(new_value).value}\nThe value of each name in the program is the following: \n{chr(10).join([f'  {name}: {db_to_string(get_value_from_namespaces(name, namespaces)).value}' for name in names])}\nThe expression used to get this value is: \n{expr.to_string()}\nThe expression used to get the indexes are as follows: \n{(chr(10) * 2).join([ex.to_string(1) for ex in index_exprs])}", statement.name)

    visited_whens = []
    if indexes:

        # goes down the list until it can assign something in the list
        def assign_variable_helper(value_to_modify: Value, remaining_indexes: list[Value]):
            if not value_to_modify or not isinstance(value_to_modify, DreamberdIndexable):
                raise_error_at_line(filename, code, name_token.line, "Attempted to index into an un-indexable object.")
            index = remaining_indexes.pop(0) 

            # check for some watchers here too!!!!!!!!!!!
            when_watchers = get_code_from_when_statement_watchers(id(value_to_modify), when_statement_watchers)
            for when_watcher in when_watchers:  # i just wanna be done with this :(
                if any([when_watcher == x for x in visited_whens]): 
                    continue
                condition, inside_statements = when_watcher
                condition_val = evaluate_expression(condition, namespaces, async_statements, when_statement_watchers)
                execute_conditional(condition_val, inside_statements, namespaces, when_statement_watchers)
                visited_whens.append(when_watcher)

            if not remaining_indexes:  # perform actual assignment here
                value_to_modify.assign_index(index, new_value)
            else:
                assign_variable_helper(value_to_modify.access_index(index), remaining_indexes)

        if isinstance(var, Variable) and not var.can_edit_value:
            raise_error_at_token(filename, code, "Cannot edit the value of this variable.", name_token)
        assign_variable_helper(var.value, indexes)
               
    else: 
        if not isinstance(var, Variable):
            raise_error_at_token(filename, code, "Attempted to set name that is not a variable.", name_token)
        if not var.can_be_reset:
            raise_error_at_token(filename, code, "Attempted to set a variable that cannot be set.", name_token)
        var.add_lifetime(new_value, confidence, 100000000000, var.can_be_reset, var.can_edit_value)

    # check if there is anything watching this value
    watchers_key = (name.split('.')[-1], id(ns))  # this shit should be a seperate function
    if watcher := name_watchers.get(watchers_key):
        st, stored_nexts, watcher_ns, promise = watcher
        mod_name = get_modified_next_name(*watchers_key)
        watcher_ns[-1][mod_name] = Name(mod_name, new_value)  # add the value to the uppermost namespace
        stored_nexts.remove(watchers_key)                   # remove the name from the set containing remaining names
        if not stored_nexts:  # not waiting on anybody else, execute the code
            interpret_name_watching_statement(st, watcher_ns, promise, async_statements, when_statement_watchers)
        del name_watchers[watchers_key]             # stop watching this name  

    # get new watchers for this
    when_watchers = get_code_from_when_statement_watchers(id(var), when_statement_watchers)
    for when_watcher in when_watchers:  # i just wanna be done with this :(
        if any([when_watcher == x for x in visited_whens]): 
            continue
        condition, inside_statements = when_watcher
        condition_val = evaluate_expression(condition, namespaces, async_statements, when_statement_watchers)
        if isinstance(new_value, DreamberdMutable):
            if id(new_value) not in when_statement_watchers[-1]:
                when_statement_watchers[-1][id(new_value)] = []  
            when_statement_watchers[-1][id(new_value)].append(when_watcher)  ##### remember : this is tuple so it is immutable and copied !!!!!!!!!!!!!!!!!!!!!!  # wait nvm i suick at pytghon
        if isinstance(var, Variable) and var.prev_values and isinstance(var.prev_values[-1], DreamberdMutable):   # if prev value was being observed under this statement, remove it  ??
            remove_from_when_statement_watchers(id(var.prev_values[-1]), when_watcher, when_statement_watchers)
        execute_conditional(condition_val, inside_statements, namespaces, when_statement_watchers)
        visited_whens.append(when_watcher)

def get_value_from_promise(val: DreamberdPromise) -> Value:
    if val.value is None:
        return DreamberdUndefined()
    return val.value

def get_name_from_namespaces(name: str, namespaces: list[Namespace]) -> Optional[Union[Variable, Name]]:
    """ This is called when we are sure that the value is a name. """
    if len(name_split := name.split('.')) == 1:
        for ns in reversed(namespaces):
            if (v := ns.get(name)) is not None:
                return v
    else:
        base_val = get_name_from_namespaces(name_split[0], namespaces)
        if not base_val:   # base object not found
            return None
        for other_name in name_split[1:]:
            if not isinstance(base_val.value, DreamberdNamespaceable):
                return None
            base_val = get_name_from_namespaces(other_name, [base_val.value.namespace])
            if not base_val:   # the value was not found in the namespace
                return None
        return base_val
    return None

def get_name_and_namespace_from_namespaces(name: str, namespaces: list[Namespace]) -> tuple[Optional[Union[Variable, Name]], Optional[Namespace]]:
    """ This is the same as the function defined above, except it also returns the namespace in which the name was found. """
    if len(name_split := name.split('.')) == 1:
        for ns in reversed(namespaces):
            if (v := ns.get(name)) is not None:
                return v, ns
    else:
        base_val, ns = get_name_and_namespace_from_namespaces(name_split[0], namespaces)
        if not base_val:
            return None, ns  # value doesn't exist but the namespace does
        for other_name in name_split[1:]:
            if not isinstance(base_val.value, DreamberdNamespaceable):
                return None, None 
            ns = base_val.value.namespace
            base_val = get_name_from_namespaces(other_name, [ns])
            if not base_val:  # same case as before
                return None, ns
        return base_val, ns
    return None, namespaces[-1]

def evaluate_escape_sequences(val: DreamberdString) -> DreamberdString:  # this needs only be called once per completed string
    return DreamberdString(eval(f'"{val.value.replace(f"{chr(34)}", f"{chr(92)}{chr(34)}")}"'))  # cursed string parsing

def interpret_formatted_string(val: Token, namespaces: list[Namespace], async_statements: AsyncStatements, when_statement_watchers: WhenStatementWatchers) -> DreamberdString:
    val_string = val.value
    try:
        locale.setlocale(locale.LC_ALL, locale.getlocale()[0])
        symbol: str = locale.localeconv()['currency_symbol']  # type: ignore 
    except locale.Error:
        symbol = '$'
    if not any(indeces := [val_string[i:i+len(symbol)] == symbol for i in range(len(val_string) - len(symbol))]):
        return DreamberdString(val_string)
    try:
        evaluated_values: list[tuple[str, tuple[int, int]]] = []  # [(str, (start, end))...]
        for group_start_index in [i for i in range(len(indeces)) if indeces[i]]:
            if val_string[group_start_index + len(symbol)] == '{':
                end_index = group_start_index + len(symbol)
                bracket_layers = 1
                while bracket_layers:  # if this errs, it will be caught and detected as invalid formatting
                    end_index += 1
                    if val_string[end_index] == '{':
                        bracket_layers += 1 
                    elif val_string[end_index] == '}':
                        bracket_layers -= 1
                
                # end_index is now the index containing the bracket.
                internal_tokens = db_tokenize(f"{filename}__interpolated_string", val_string[group_start_index + len(symbol) + 1 : end_index ])
                internal_expr = build_expression_tree(filename, internal_tokens, code)
                internal_value = evaluate_expression(internal_expr, namespaces, async_statements, when_statement_watchers, ignore_string_escape_sequences=True)
                evaluated_values.append((db_to_string(internal_value).value, (group_start_index, end_index + 1)))

        new_string = list(val_string)
        for replacement, (start, end) in reversed(evaluated_values):
            new_string[start:end] = replacement
        return DreamberdString(''.join(new_string))
                
    except IndexError:
        raise_error_at_line(filename, code, current_line, "Invalid interpolated string formatting.")

def determine_non_name_value(val: Token) -> Value:
    """ 
    Takes a string/Token and determines if the value is a number, string, or invalid. 
    Valid names should have been found already by the previous function.
    """
    global deleted_values
    retval = None
    if len(v := val.value.split('.')) <= 2:
        if all(x.isdigit() for x in v):
            retval = DreamberdNumber([int, float][len(v) - 1](val.value))
    if not retval:
        retval = DreamberdString(val.value)
    if retval in deleted_values:
        raise_error_at_line(filename, code, val.line, f"The value {retval.value} has been deleted.")
    return retval

def is_approx_equal(left: Value, right: Value) -> DreamberdBoolean:

    if is_really_really_equal(left, right).value:
        return DreamberdBoolean(True)

    if isinstance(left, DreamberdString) or isinstance(right, DreamberdString):

        return DreamberdBoolean(SequenceMatcher(None, db_to_string(left).value, 
                                                db_to_string(right).value).ratio() > STRING_EQUALITY_RATIO)

    if isinstance((num := left), DreamberdNumber) or isinstance((num := right), DreamberdNumber):
        other = left if right == num else right
        if isinstance(other, (DreamberdNumber, DreamberdUndefined, DreamberdBoolean)):
            left_num, right_num = db_to_number(left).value, db_to_number(right).value
            return DreamberdBoolean(left_num == right_num or (False if left_num == 0 else (left_num - right_num) / left_num < NUM_EQUALITY_RATIO))

    if isinstance(left, DreamberdBoolean) or isinstance(right, DreamberdBoolean):
        left_bool, right_bool = db_to_boolean(left).value, db_to_boolean(right).value
        if left_bool is None or right_bool is None:
            return DreamberdBoolean(None)  # maybe
        return DreamberdBoolean(left_bool == right_bool)

    if (val := db_to_boolean(left).value) == db_to_boolean(right).value and val is not None:
        return DreamberdBoolean(True)

    if type(left) != type(right):
        return DreamberdBoolean(None)  # maybe, programmer got too lazy

    if isinstance(left, DreamberdList) and isinstance(right, DreamberdList):
        if len(left.values) == len(right.values) == 0:
            return DreamberdBoolean(True)
        is_equals = [is_approx_equal(l, r) for l, r in zip(left.values, right.values)]
        ratio = sum([int(x.value) if x.value is not None else 0.5 for x in is_equals]) / max(len(left.values), len(right.values))
        return DreamberdBoolean(ratio > LIST_EQUALITY_RATIO)

    if isinstance(left, DreamberdMap) and isinstance(right, DreamberdMap):
        if len(left.self_dict) == len(right.self_dict) == 0:
            return DreamberdBoolean(True)
        is_equals = [is_approx_equal(left.self_dict[key], right.self_dict[key])
                     for key in left.self_dict.keys() & right.self_dict.keys()]
        ratio = sum([int(x.value) if x.value is not None else 0.5 for x in is_equals]) /                  \
                len(left.self_dict.keys() | right.self_dict.keys())
        return DreamberdBoolean(ratio > MAP_EQUALITY_RATIO)

    if isinstance(left, DreamberdFunction) and isinstance(right, DreamberdFunction):
        if len(left.code) == len(right.code) == 0:
            return DreamberdBoolean(True)
        ratio = sum([len(set(l) | set(r)) / min(len(l), len(r)) for l, r in zip(left.code, right.code)]) / max(len(left.code), len(right.code))
        return DreamberdBoolean(True if ratio > FUNCTION_EQUALITY_RATIO else None)  # for no reason whatsoever, this will be maybe and not False

    if isinstance(left, DreamberdObject) and isinstance(right, DreamberdObject):
        if len(left.namespace) == len(right.namespace) == 0:
            return DreamberdBoolean(True)
        is_equals = [is_approx_equal(left.namespace[key].value, right.namespace[key].value)
                     for key in left.namespace.keys() & right.namespace.keys()]
        ratio = sum([int(x.value) if x.value is not None else 0.5 for x in is_equals]) /                  \
                len(left.namespace.keys() | right.namespace.keys())
        return DreamberdBoolean(ratio > OBJECT_EQUALITY_RATIO)

    return DreamberdBoolean(None)

def is_equal(left: Value, right: Value) -> DreamberdBoolean:

    if isinstance(left, DreamberdString) or isinstance(right, DreamberdString):
        return DreamberdBoolean(db_to_string(left).value == db_to_string(right).value)

    if isinstance(left, DreamberdNumber) or isinstance(right, DreamberdNumber):
        return DreamberdBoolean(db_to_number(left).value == db_to_number(right).value)

    if isinstance(left, DreamberdBoolean) or isinstance(right, DreamberdBoolean):
        left_bool, right_bool = db_to_boolean(left).value, db_to_boolean(right).value
        if left_bool is None or right_bool is None:
            return DreamberdBoolean(None)  # maybe
        return DreamberdBoolean(left_bool == right_bool)

    if (val := db_to_boolean(left).value) == db_to_boolean(right).value and val is not None:
        return DreamberdBoolean(True)

    if type(left) != type(right):
        return DreamberdBoolean(None)  # maybe, programmer got too lazy

    if isinstance(left, DreamberdList) and isinstance(right, DreamberdList):
        return DreamberdBoolean(all([is_equal(l, r).value for l, r in zip(left.values, right.values)]))

    if isinstance(left, DreamberdMap) and isinstance(right, DreamberdMap):
        is_equals = [is_approx_equal(left.self_dict[key], right.self_dict[key]).value 
                     for key in left.self_dict.keys() & right.self_dict.keys()]
        return DreamberdBoolean(all(is_equals))

    if isinstance(left, DreamberdObject) and isinstance(right, DreamberdObject):
        is_equals = [is_approx_equal(left.namespace[key].value, right.namespace[key].value).value 
                     for key in left.namespace.keys() & right.namespace.keys()]
        return DreamberdBoolean(all(is_equals))

    return DreamberdBoolean(None)

def is_really_equal(left: Value, right: Value) -> DreamberdBoolean:
    if type(left) != type(right):
        return DreamberdBoolean(False)
    match left, right:  # i know these are horribly verbose but if i don't do this my LSP yells at me
        case (DreamberdNumber(), DreamberdNumber()) | \
             (DreamberdString(), DreamberdString()) | \
             (DreamberdBoolean(), DreamberdBoolean()) | \
             (DreamberdKeyword(), DreamberdKeyword()): 
            return DreamberdBoolean(left.value == right.value)
        case (DreamberdUndefined(), DreamberdUndefined()):
            return DreamberdBoolean(True) 
        case (DreamberdObject(), DreamberdObject()):
            return DreamberdBoolean(left.class_name == right.class_name and 
                                    left.namespace.keys() == right.namespace.keys() and 
                                    all([is_really_equal(left.namespace[k].value, right.namespace[k].value).value for k in left.namespace.keys()]))
        case (DreamberdFunction(), DreamberdFunction()):
            return DreamberdBoolean(all([getattr(left, name) == getattr(right, name) for name in ["code", "args", "is_async"]]))
        case (DreamberdList(), DreamberdList()):
            return DreamberdBoolean(len(left.values) == len(right.values) and 
                                    all([is_really_equal(l, r).value for l, r in zip(left.values, right.values)]))
        case (DreamberdMap(), DreamberdMap()):
            return DreamberdBoolean(left.self_dict.keys() == right.self_dict.keys() and 
                                    all([is_really_equal(left.self_dict[k], right.self_dict[k]).value for k in left.self_dict]))
    return DreamberdBoolean(None)

def is_really_really_equal(left: Value, right: Value) -> DreamberdBoolean:
    return DreamberdBoolean(left is right)

def is_less_than(left: Value, right: Value) -> DreamberdBoolean:
    if type(left) != type(right):
        raise_error_at_line(filename, code, current_line, f"Cannot compare value of type {type(left).__name__} with one of type {type(right).__name__}.")
    match left, right:
        case (DreamberdNumber(), DreamberdNumber()) | \
             (DreamberdString(), DreamberdString()) | \
             (DreamberdBoolean(), DreamberdBoolean()):
            if isinstance(left, DreamberdBoolean) and isinstance(right, DreamberdBoolean) and (left.value is None or right.value is None):
                return DreamberdBoolean(None)
            return DreamberdBoolean(left.value < right.value)   # type: ignore
        case (DreamberdUndefined(), DreamberdUndefined()):
            return DreamberdBoolean(False)
        case (DreamberdList(), DreamberdList()):
            return DreamberdBoolean(len(left.values) < len(right.values))
        case (DreamberdMap(), DreamberdMap()):
            return DreamberdBoolean(len(left.self_dict) < len(right.self_dict))
        case (DreamberdKeyword(), DreamberdKeyword()) | \
             (DreamberdObject(), DreamberdObject()) | \
             (DreamberdFunction(), DreamberdFunction()):
            raise_error_at_line(filename, code, current_line, f"Comparison not supported between elements of type {type(left).__name__}.")
    return DreamberdBoolean(None)

def perform_single_value_operation(val: Value, operator_token: Token) -> Value: 
    match operator_token.type:
        case TokenType.SUBTRACT:
            match val:
                case DreamberdNumber():
                    return DreamberdNumber(-val.value)
                case DreamberdList():
                    return DreamberdList(val.values[::-1])
                case DreamberdString():
                    return DreamberdString(val.value[::-1])
                case _:
                    raise_error_at_token(filename, code, f"Cannot negate a value of type {type(val).__name__}", operator_token)
        case TokenType.SEMICOLON:
            val_bool = db_to_boolean(val)
            return db_not(val_bool) 
    raise_error_at_token(filename, code, "Something went wrong. My bad.", operator_token)

def perform_two_value_operation(left: Value, right: Value, operator: OperatorType, operator_token: Token) -> Value:
    match operator:
        case OperatorType.ADD:
            if isinstance(left, DreamberdString) or isinstance(right, DreamberdString):
                return DreamberdString(db_to_string(left).value + db_to_string(right).value)
            left_num = db_to_number(left)
            right_num = db_to_number(right)
            return DreamberdNumber(left_num.value + right_num.value)
        case OperatorType.SUB | OperatorType.MUL | OperatorType.DIV | OperatorType.EXP:
            left_num = db_to_number(left)
            right_num = db_to_number(right)
            if operator == OperatorType.DIV and abs(right_num.value) < FLOAT_TO_INT_PREC: # pretty much zero
                return DreamberdUndefined() 
            elif operator == OperatorType.EXP and left_num.value < -FLOAT_TO_INT_PREC and not is_int(right_num.value):
                raise_error_at_line(filename, code, current_line, "Cannot raise a negative base to a non-integer exponent.")
            match operator:
                case OperatorType.SUB: result = left_num.value - right_num.value
                case OperatorType.MUL: result = left_num.value * right_num.value
                case OperatorType.DIV: result = left_num.value / right_num.value
                case OperatorType.EXP: result = pow(left_num.value, right_num.value)
            return DreamberdNumber(result)
        case OperatorType.OR:
            left_bool = db_to_boolean(left)
            right_bool = db_to_boolean(right) 
            match left_bool.value, right_bool.value:
                case True, _:     return left    # yes 
                case False, _:    return right   # depends
                case None, True:  return right   # yes
                case None, False: return left    # maybe?
                case None, None:  return left if random.random() < 0.50 else right   # maybe? 
        case OperatorType.AND:  
            left_bool = db_to_boolean(left)
            right_bool = db_to_boolean(right) 
            match left_bool.value, right_bool.value:
                case True, _:     return right   # depends
                case False, _:    return left    # nope
                case None, True:  return left    # maybe?
                case None, False: return right   # nope
                case None, None:  return left if random.random() < 0.50 else right  # maybe? 
        case OperatorType.E: 
            return is_approx_equal(left, right)

        # i'm gonna call this lasagna code because it's stacked like lasagna and looks stupid
        case OperatorType.EE | OperatorType.NE:
            if operator == OperatorType.EE:
                return is_equal(left, right)
            return db_not(is_equal(left, right))
        case OperatorType.EEE | OperatorType.NEE:
            if operator == OperatorType.EEE:
                return is_really_equal(left, right)
            return db_not(is_really_equal(left, right))
        case OperatorType.EEEE | OperatorType.NEEE:
            if operator == OperatorType.EEEE:
                return is_really_really_equal(left, right)
            return db_not(is_really_really_equal(left, right))
        case OperatorType.GT | OperatorType.LE:
            is_eq = is_really_equal(left, right)
            is_less = is_less_than(left, right)
            is_le = False
            match is_eq.value, is_less.value:  # performs the OR operation
                case (True, _) | (_, True): is_le = True
                case (None, _) | (_, None): is_le = None 
            if operator == OperatorType.LE:
                return DreamberdBoolean(is_le)
            return db_not(DreamberdBoolean(is_le))
        case OperatorType.LT | OperatorType.GE:  
            if operator == OperatorType.LT:
                return is_less_than(left, right)
            return db_not(is_less_than(left, right))

    raise_error_at_token(filename, code, "Something went wrong here.", operator_token)

def get_value_from_namespaces(name_or_value: Token, namespaces: list[Namespace]) -> Value:

    # what the frick am i doing rn
    if v := get_name_from_namespaces(name_or_value.value, namespaces):
        if isinstance(v.value, DreamberdPromise):
            return deepcopy(get_value_from_promise(v.value))  # consider not deepcopying this but it doesnt really matter
        return v.value
    return determine_non_name_value(name_or_value)

def print_expression_debug(debug: int, expr: Union[list[Token], ExpressionTreeNode], value: Value, namespaces: list[Namespace]) -> None:
    expr = get_built_expression(expr)
    msg = None
    match debug:
        case 0: pass 
        case 1: 
            msg = f"Expression evaluates to value {db_to_string(value).value}."
        case 2: 
            names = gather_names_or_values(expr)
            msg = f"Expression evaluates to value {db_to_string(value).value}.\nThe value of each name in the expression is the following: \n{chr(10).join([f'  {name}: {db_to_string(get_value_from_namespaces(name, namespaces)).value}' for name in names])}"
        case _: 
            names = gather_names_or_values(expr)
            msg = f"Expression evaluates to value {db_to_string(value).value}.\nThe value of each name in the expression is the following: \n{chr(10).join([f'  {name}: {db_to_string(get_value_from_namespaces(name, namespaces)).value}' for name in names])}\nThe expression used to get this value is: \n{expr.to_string()}"

    if not msg: return
    if t := get_expr_first_token(expr):
        debug_print(filename, code, msg, t)
    else: debug_print_no_token(filename, msg)


def evaluate_expression(expr: Union[list[Token], ExpressionTreeNode], namespaces: list[dict[str, Union[Variable, Name]]], async_statements: AsyncStatements, when_statement_watchers: WhenStatementWatchers, *, ignore_string_escape_sequences: bool = False) -> Value:
    """ Wrapper for the evaluate_expression_for_real function that checks deleted values on each run. """
    retval = evaluate_expression_for_real(expr, namespaces, async_statements, when_statement_watchers, ignore_string_escape_sequences)
    if isinstance(retval, (DreamberdNumber, DreamberdString)) and retval in deleted_values:
        raise_error_at_line(filename, code, current_line, f"The value {retval.value} has been deleted.")
    return retval

def evaluate_expression_for_real(expr: Union[list[Token], ExpressionTreeNode], namespaces: list[dict[str, Union[Variable, Name]]], async_statements: AsyncStatements, when_statement_watchers: WhenStatementWatchers, ignore_string_escape_sequences: bool) -> Value:

    expr = get_built_expression(expr)
    match expr:
        case FunctionNode():  # done :)
            
            # for a function, the thing must be in the namespace
            func = get_name_from_namespaces(expr.name.value, namespaces)

            # make sure it exists and it is actually a function in the namespace
            if func is None:
                raise_error_at_token(filename, code, "Cannot find token in namespace.", expr.name)
    
            # check the thing in the await symbol. if awaiting a single function that is async, evaluate it as not async
            force_execute_sync = False
            if isinstance(func.value, DreamberdKeyword):
                if func.value.value == "await":
                    if len(expr.args) != 1:
                        raise_error_at_token(filename, code, "Expected only one argument for await function.", expr.name)
                    if not isinstance(expr.args[0], FunctionNode):
                        raise_error_at_token(filename, code, "Expected argument of await function to be a function call.", expr.name)
                    force_execute_sync = True 
                    
                    # check for None again
                    expr = expr.args[0]
                    func = get_name_from_namespaces(expr.name.value, namespaces)
                    if func is None:  # the other check happens in the next statement
                        raise_error_at_token(filename, code, "Cannot find token in namespaces.", expr.name)

                elif func.value.value == "previous":
                    if len(expr.args) != 1:
                        raise_error_at_token(filename, code, "Expected only one argument for previous function.", expr.name)
                    if not isinstance(expr.args[0], ValueNode):
                        raise_error_at_token(filename, code, "Expected argument of previous function to be a variable.", expr.name)
                    force_execute_sync = True 
                    
                    # check for None again
                    val = get_name_from_namespaces(expr.args[0].name_or_value.value, namespaces)
                    if not isinstance(val, Variable):
                        raise_error_at_token(filename, code, "Expected argument of previous function to be a defined variable.", expr.args[0].name_or_value)
                    return val.prev_values[-1]

            if not isinstance(func.value, (BuiltinFunction, DreamberdFunction)):
                raise_error_at_token(filename, code, "Attempted function call on non-function value.", expr.name)
            
            caller = None
            if len(name_split := expr.name.value.split('.')) > 1:
                caller = '.'.join(name_split[:-1])
                expr = deepcopy(expr)   # we create a copy of the expression as to not modify it badly
                expr.args.insert(0, ValueNode(Token(TokenType.NAME, caller, expr.name.line, expr.name.col)))  # artificially put this here, as this is the imaginary "this" 
            args = [evaluate_expression(arg, namespaces, async_statements, when_statement_watchers) for arg in expr.args]
            if isinstance(args[0], DreamberdSpecialBlankValue):
                args = args[1:]
            if isinstance(func.value, DreamberdFunction) and func.value.is_async and not force_execute_sync:
                register_async_function(expr, func.value, namespaces, args, async_statements)
                return DreamberdUndefined()
            elif isinstance(func.value, BuiltinFunction) and func.value.modifies_caller:  # special cases where the function itself modifies the caller
                if caller:  # seems like a needless check but it makes the errors go away
                    caller_var = get_name_from_namespaces(caller, namespaces)
                    if isinstance(caller_var, Variable) and not caller_var.can_edit_value:
                        raise_error_at_line(filename, code, current_line, "Cannot edit the value of this variable.")

                retval = evaluate_normal_function(expr, func.value, namespaces, args, when_statement_watchers)
                when_watchers = get_code_from_when_statement_watchers(id(args[0]), when_statement_watchers)
                for when_watcher in when_watchers:  # i just wanna be done with this :(
                    condition, inside_statements = when_watcher
                    condition_val = evaluate_expression(condition, namespaces, async_statements, when_statement_watchers)
                    execute_conditional(condition_val, inside_statements, namespaces, when_statement_watchers)
                return retval

            return evaluate_normal_function(expr, func.value, namespaces, args, when_statement_watchers)

        case ListNode():  # done :) 
            return DreamberdList([evaluate_expression(x, namespaces, async_statements, when_statement_watchers) for x in expr.values])

        case ValueNode():  # done :)
            if expr.name_or_value.type == TokenType.STRING: 
                retval = interpret_formatted_string(expr.name_or_value, namespaces, async_statements, when_statement_watchers)
                if not ignore_string_escape_sequences:
                    return evaluate_escape_sequences(retval)
                return retval
            return get_value_from_namespaces(expr.name_or_value, namespaces)

        case IndexNode():  # done :)
            value = evaluate_expression(expr.value, namespaces, async_statements, when_statement_watchers)
            index = evaluate_expression(expr.index, namespaces, async_statements, when_statement_watchers)
            if not isinstance(value, DreamberdIndexable):
                raise_error_at_line(filename, code, current_line, "Attempting to index a value that is not indexable.")
            return value.access_index(index)

        case ExpressionNode():  # done :)
            left = evaluate_expression(expr.left, namespaces, async_statements, when_statement_watchers)
            if db_to_boolean(left).value == True and expr.operator == OperatorType.OR:   # handle short curcuiting for True or __
                return left
            elif db_to_boolean(left).value == False and expr.operator == OperatorType.AND:   # handle short curcuiting for False and __
                return left
            right = evaluate_expression(expr.right, namespaces, async_statements, when_statement_watchers)
            return perform_two_value_operation(left, right, expr.operator, expr.operator_token)

        case SingleOperatorNode():
            val = evaluate_expression(expr.expression, namespaces, async_statements, when_statement_watchers)
            return perform_single_value_operation(val, expr.operator)

    return DreamberdUndefined()

def handle_next_expressions(expr: ExpressionTreeNode, namespaces: list[Namespace]) -> tuple[ExpressionTreeNode, set[tuple[str, int]], set[str]]:

    """ 
    This function looks for the "next" keyword in an expression, and detects seperate await modifiers for that keyword.
    Then, it removes the "next" and "await next" nodes from the ExpressionTree, and returns the head of the tree.
    Additionally, every name that appears in the function as a next or async next, its value is saved in a temporary namespace.
    With the returned set of names that are used in "next" and "await next", we can insert these into a dictionary
        that contains information about which names are being "watched" for changes. When this dictionary changes,
        we can execute code accordingly.

    This is my least favorite function in the program.
    """

    normal_nexts: set[tuple[str, int]] = set()
    async_nexts: set[str] = set()
    inner_nexts: list[tuple[set[tuple[str, int]], set[str]]] = []
    match expr:
        case FunctionNode():

            func = get_name_from_namespaces(expr.name.value, namespaces)
            if func is None:
                raise_error_at_token(filename, code, "Attempted function call on undefined variable.", expr.name)

            # check if it is a next or await 
            is_next = is_await = False   # i don't need this but it makes my LSP stop crying so it's here
            if isinstance(func.value, DreamberdKeyword) and \
               ((is_next := func.value.value == "next") or (is_await := func.value.value == "await")):

                if is_next:

                    # add it to list of things to watch for and change the returned expression to the name being next-ed
                    if len(expr.args) != 1 or not isinstance(expr.args[0], ValueNode):
                        raise_error_at_token(filename, code, "\"Next\"keyword can only take a single value as an argument.", expr.name)
                    name = expr.args[0].name_or_value.value
                    _, ns = get_name_and_namespace_from_namespaces(name, namespaces)
                    if not ns:
                        raise_error_at_line(filename, code, current_line, "Attempted to access namespace of a value without a namespace.")
                    last_name = name.split('.')[-1]
                    normal_nexts.add((name, id(ns)))
                    expr = expr.args[0]
                    expr.name_or_value.value = get_modified_next_name(last_name, id(ns))

                elif is_await:

                    if len(expr.args) != 1 or not isinstance(expr.args[0], FunctionNode):
                        raise_error_at_token(filename, code, "Can only await a function.", expr.name)
                    inner_expr = expr.args[0]
                        
                    func = get_name_from_namespaces(expr.args[0].name.value, namespaces)
                    if func is None:
                        raise_error_at_token(filename, code, "Attempted function call on undefined variable.", expr.name)

                    if isinstance(func.value, DreamberdKeyword) and func.value.value == "next":
                        if len(inner_expr.args) != 1 or not isinstance(inner_expr.args[0], ValueNode):
                            raise_error_at_token(filename, code, "\"Next\"keyword can only take a single value as an argument.", inner_expr.name)
                        name = inner_expr.args[0].name_or_value.value 
                        _, ns = get_name_and_namespace_from_namespaces(name, namespaces)
                        if not ns:
                            raise_error_at_line(filename, code, current_line, "Attempted to access namespace of a value without a namespace.")
                        last_name = name.split('.')[-1]
                        async_nexts.add(name)  # only need to store the name for the async ones because we are going to wait anyways
                        expr = inner_expr.args[0]
                        expr.name_or_value.value = get_modified_next_name(last_name, id(ns))

            else:
                replacement_args = []
                for arg in expr.args:
                    new_expr, normal_arg_nexts, async_arg_nexts = handle_next_expressions(arg, namespaces)
                    inner_nexts.append((normal_arg_nexts, async_arg_nexts))
                    replacement_args.append(new_expr)
                expr.args = replacement_args
                
        case ListNode():
            replacement_values = []
            for ex in expr.values:
                new_expr, normal_expr_nexts, async_expr_nexts = handle_next_expressions(ex, namespaces)
                inner_nexts.append((normal_expr_nexts, async_expr_nexts))
                replacement_values.append(new_expr)
            expr.values = replacement_values
        case IndexNode():
            new_value, normal_value_nexts, async_value_nexts = handle_next_expressions(expr.value, namespaces)
            new_index, normal_index_nexts, async_index_nexts = handle_next_expressions(expr.index, namespaces)
            expr.value = new_value 
            expr.index = new_index
            inner_nexts.extend([(normal_value_nexts, async_value_nexts), (normal_index_nexts, async_index_nexts)])
        case ExpressionNode():
            new_left, normal_left_nexts, async_left_nexts = handle_next_expressions(expr.left, namespaces)
            new_right, normal_right_nexts, async_right_nexts = handle_next_expressions(expr.right, namespaces)
            expr.left = new_left 
            expr.right = new_right
            inner_nexts.extend([(normal_left_nexts, async_left_nexts), (normal_right_nexts, async_right_nexts)])
        case SingleOperatorNode():
            new_expr, normal_expr_nexts, async_expr_nexts = handle_next_expressions(expr.expression, namespaces)
            expr.expression = new_expr
            inner_nexts.append((normal_expr_nexts, async_expr_nexts))
    for nn, an in inner_nexts:
        normal_nexts |= nn 
        async_nexts |= an 
    return expr, normal_nexts, async_nexts

def save_previous_values_next_expr(expr_to_modify: ExpressionTreeNode, nexts: set[str], namespaces: list[Namespace]) -> Namespace:

    saved_namespace: Namespace = {}
    match expr_to_modify:
        case ValueNode():
            if expr_to_modify.name_or_value.type == TokenType.STRING:
                return {}
            name = expr_to_modify.name_or_value.value
            if name not in nexts:
                return {}
            val = get_name_from_namespaces(name, namespaces)
            if not val:
                val = Name("", determine_non_name_value(expr_to_modify.name_or_value))
            mod_name = get_modified_prev_name(name)
            expr_to_modify.name_or_value.value = mod_name
            return {mod_name: Name(mod_name, val.value)}
        case ExpressionNode():
            left_ns = save_previous_values_next_expr(expr_to_modify.left, nexts, namespaces)
            right_ns = save_previous_values_next_expr(expr_to_modify.right, nexts, namespaces)
            return left_ns | right_ns
        case IndexNode():
            value_ns = save_previous_values_next_expr(expr_to_modify.value, nexts, namespaces)
            index_ns = save_previous_values_next_expr(expr_to_modify.index, nexts, namespaces)
            return value_ns | index_ns
        case ListNode():
            for ex in expr_to_modify.values:
                saved_namespace |= save_previous_values_next_expr(ex, nexts, namespaces)
            return saved_namespace
        case FunctionNode():
            for arg in expr_to_modify.args:
                saved_namespace |= save_previous_values_next_expr(arg, nexts, namespaces)
            return saved_namespace
        case SingleOperatorNode():
            return save_previous_values_next_expr(expr_to_modify.expression, nexts, namespaces)
    return saved_namespace

def determine_statement_type(possible_statements: tuple[CodeStatement, ...], namespaces: list[Namespace]) -> Optional[CodeStatement]:
    instance_to_keywords: dict[type[CodeStatementKeywordable], set[str]] = {
        Conditional: {'if'},
        WhenStatement: {'when'},
        AfterStatement: {'after'},
        ClassDeclaration: {'class', 'className'},
        DeleteStatement: {'delete'},
        ReverseStatement: {'reverse'},
        ImportStatement: {'import'}
    }

    for st in possible_statements:
        if isinstance(st, CodeStatementKeywordable):
            val = get_name_from_namespaces(st.keyword.value, namespaces)
            if val is not None and isinstance(val.value, DreamberdKeyword) and val.value.value in instance_to_keywords[type(st)]:
                return st
        elif isinstance(st, ReturnStatement):
            if st.keyword is None:
                return st 
            val = get_name_from_namespaces(st.keyword.value, namespaces)
            if val and isinstance(val.value, DreamberdKeyword) and val.value.value == "return":
                return st 
        elif isinstance(st, FunctionDefinition):  # allow for async and normal function definitions
            if len(st.keywords) == 1:
                val = get_name_from_namespaces(st.keywords[0].value, namespaces)
                if val and isinstance(val.value, DreamberdKeyword) and re.match(r"^f?u?n?c?t?i?o?n?$", val.value.value):
                    return st
            elif len(st.keywords) == 2:
                val = get_name_from_namespaces(st.keywords[0].value, namespaces)
                other_val = get_name_from_namespaces(st.keywords[1].value, namespaces)
                if val and other_val and isinstance(val.value, DreamberdKeyword) and isinstance(other_val.value, DreamberdKeyword) \
                   and re.match(r"^f?u?n?c?t?i?o?n?$", other_val.value.value) and val.value.value == 'async':
                    return st
        elif isinstance(st, VariableDeclaration):  # allow for const const const and normal declarations
            if len(st.modifiers) == 2:
                if all([(val := get_name_from_namespaces(mod.value, namespaces)) is not None and 
                    isinstance(val.value, DreamberdKeyword) and val.value.value in {'const', 'var'}
                    for mod in st.modifiers]):
                    return st
            elif len(st.modifiers) == 3:
                if all([(val := get_name_from_namespaces(mod.value, namespaces)) is not None and 
                    isinstance(val.value, DreamberdKeyword) and val.value.value == 'const' 
                    for mod in st.modifiers]):
                    return st
        elif isinstance(st, ExportStatement):
            if isinstance(v := get_value_from_namespaces(st.to_keyword, namespaces), DreamberdKeyword) and v.value == 'to' and \
               isinstance(v := get_value_from_namespaces(st.export_keyword, namespaces), DreamberdKeyword) and v.value == 'export':
                return st

    # now is left: expression evalulation and variable assignment
    for st in possible_statements:
        if isinstance(st, VariableAssignment):
            return st 
    for st in possible_statements:
        if isinstance(st, ExpressionStatement):
            return st
    return None 

def adjust_for_normal_nexts(statement: CodeStatementWithExpression, async_nexts: set[str], normal_nexts: set[tuple[str, int]], promise: Optional[DreamberdPromise], namespaces: list[Namespace], prev_namespace: Namespace):

    old_async_vals, old_normal_vals = [], []
    get_state_watcher = lambda val: None if not val else len(v) if (v := getattr(val, "prev_values")) else 0
    for name in async_nexts:
        old_async_vals.append(get_state_watcher(get_name_from_namespaces(name, namespaces))) 
    for name, _ in normal_nexts:
        old_normal_vals.append(get_state_watcher(get_name_from_namespaces(name, namespaces)))

    # for each async one, wait until each one is different
    for name, start_len in zip(async_nexts, old_async_vals): 
        curr_len = get_state_watcher(get_name_from_namespaces(name, namespaces))
        while start_len != curr_len:  
            curr_len = get_state_watcher(get_name_from_namespaces(name, namespaces))

    # now, build a namespace for each one
    new_namespace: Namespace = {}
    for name, old_len in zip(async_nexts, old_async_vals):
        v, ns = get_name_and_namespace_from_namespaces(name, namespaces)
        if not v or not ns or (old_len is not None and not isinstance(v, Variable)):
            raise_error_at_line(filename, code, current_line, "Something went wrong with accessing the next value of a variable.")
        mod_name = get_modified_next_name(name, id(ns))
        match old_len:
            case None: new_namespace[mod_name] = Name(mod_name, v.value if isinstance(v, Name) else v.prev_values[0])
            case i:
                if not isinstance(v, Variable):
                    raise_error_at_line(filename, code, current_line, "Something went wrong.")
                new_namespace[mod_name] = Name(mod_name, v.prev_values[i])

    # now, adjust for any values that may have already been modified by next statements 
    for (name, ns_id), old_len in zip(normal_nexts, old_normal_vals):
        new_len = get_state_watcher(v := get_name_from_namespaces(name, namespaces))
        if v is None or new_len == old_len: 
            continue
        mod_name = get_modified_next_name(name, ns_id)
        normal_nexts.remove((name, ns_id))
        match old_len:
            case None: new_namespace[mod_name] = Name(mod_name, v.value if isinstance(v, Name) else v.prev_values[0])
            case i:
                if not isinstance(v, Variable):
                    raise_error_at_line(filename, code, current_line, "Something went wrong.")
                new_namespace[mod_name] = Name(mod_name, v.prev_values[i])

    # the remaining values are still waiting on a result, add these to the list of name watchers
    # this new_namespace i am adding is purely for use in evaluation of expressions, and the code within 
    # a code statement should not use that namespace. therefore, some logic must be done to remove that namespace 
    # when the expression of a code statement is executed
    for name, ns_id in normal_nexts:
        name_watchers[(name.split('.')[-1], ns_id)] = (statement, normal_nexts, namespaces + [new_namespace | prev_namespace], promise)

def wait_for_async_nexts(async_nexts: set[str], namespaces: list[Namespace]) -> Namespace:

    old_async_vals = []
    get_state_watcher = lambda val: None if not val else len(v) if (v := getattr(val, "prev_values")) else 0
    for name in async_nexts:
        old_async_vals.append(get_state_watcher(get_name_from_namespaces(name, namespaces))) 

    # for each async one, wait until each one is different
    for name, start_len in zip(async_nexts, old_async_vals): 
        curr_len = get_state_watcher(get_name_from_namespaces(name, namespaces))
        while start_len != curr_len:  
            curr_len = get_state_watcher(get_name_from_namespaces(name, namespaces))

    # now, build a namespace for each one
    new_namespace: Namespace = {}
    for name, old_len in zip(async_nexts, old_async_vals):
        v, ns = get_name_and_namespace_from_namespaces(name, namespaces)
        if not v or not ns or (old_len is not None and not isinstance(v, Variable)):
            raise_error_at_line(filename, code, current_line, "Something went wrong with accessing the next value of a variable.") 
        mod_name = get_modified_next_name(name, id(ns))
        match old_len:
            case None: new_namespace[mod_name] = Name(mod_name, v.value if isinstance(v, Name) else v.prev_values[0])
            case i:
                if not isinstance(v, Variable):
                    raise_error_at_line(filename, code, current_line, "Something went wrong.")
                new_namespace[mod_name] = Name(mod_name, v.prev_values[i])
    return new_namespace

def interpret_name_watching_statement(statement: CodeStatementWithExpression, namespaces: list[Namespace], promise: Optional[DreamberdPromise], async_statements: AsyncStatements, when_statement_watchers: WhenStatementWatchers): 

    # evaluate the expression using the names off the top
    expr_val = evaluate_expression(statement.expression, namespaces, async_statements, when_statement_watchers) 
    index_vals = [evaluate_expression(expr, namespaces, async_statements, when_statement_watchers) 
                  for expr in statement.indexes] if isinstance(statement, VariableAssignment) else []
    namespaces.pop()  # remove expired namespace  -- THIS IS INCREDIBLY IMPORTANT

    match statement:
        case ReturnStatement():
            if promise is None:
                raise_error_at_line(filename, code, current_line, "Something went wrong.")
            promise.value = expr_val  # simply change the promise to that value as the return statement already returned a promise
        case VariableDeclaration():
            declare_new_variable(statement, expr_val, namespaces, async_statements, when_statement_watchers)
        case VariableAssignment():
            assign_variable(statement, index_vals, expr_val, namespaces, async_statements, when_statement_watchers)
        case Conditional():
            execute_conditional(expr_val, statement.code, namespaces, when_statement_watchers)
        case AfterStatement():
            execute_after_statement(expr_val, statement.code, namespaces, when_statement_watchers)
        case ExpressionStatement(): 
            print_expression_debug(statement.debug, statement.expression, expr_val, namespaces)

def clear_temp_namespace(namespaces: list[Namespace], temp_namespace: Namespace) -> None:
    for key in temp_namespace:
        del namespaces[-1][key]

# simply execute the conditional inside a new scope
def execute_conditional(condition: Value, statements_inside_scope: list[tuple[CodeStatement, ...]], namespaces: list[Namespace], when_statement_watchers: WhenStatementWatchers) -> Optional[Value]:
    condition = db_to_boolean(condition)
    execute = condition.value == True if condition.value is not None else random.random() < 0.50
    if execute:  
        return interpret_code_statements(statements_inside_scope, namespaces + [{}], [], when_statement_watchers + [{}]) # empty scope and async statements, just for this :)

# this is the equaivalent of an event listener
def get_mouse_event_object(x: int, y: int, button: mouse.Button, event: str) -> DreamberdObject:
    return DreamberdObject("MouseEvent", {
        'x': Name('x', DreamberdNumber(x)),
        'y': Name('y', DreamberdNumber(y)),
        'button': Name('button', DreamberdString(str(button).split('.')[-1])),
        'event': Name('event', DreamberdString(event)),
    })

def get_keyboard_event_object(key: Optional[Union[keyboard.Key, keyboard.KeyCode]], event: str) -> DreamberdObject:
    return DreamberdObject("MouseEvent", {
        'key': Name('key', DreamberdString(str(key).split('.')[-1])),
        'event': Name('event', DreamberdString(event)),
    })

def execute_after_statement(event: Value, statements_inside_scope: list[tuple[CodeStatement, ...]], namespaces: list[Namespace], when_statement_watchers: WhenStatementWatchers) -> None:

    if not KEY_MOUSE_IMPORTED:
        raise_error_at_line(filename, code, current_line, "Attempted to use mouse and keyboard functionality without importing the [input] extra dependency.")

    if not isinstance(event, DreamberdString):
        raise_error_at_line(filename, code, current_line, f"Invalid event for the \"after\" statement: \"{db_to_string(event)}\"")

    match event.value:
        case "mouseclick":
            mouse_buttons = {}
            def listener_func(x: int, y: int, button: mouse.Button, pressed: bool):
                nonlocal namespaces, statements_inside_scope
                if pressed:
                    mouse_buttons[button] = (x, y)
                else: 
                    if mouse_buttons[button]:   # it has been released and then pressed again
                        interpret_code_statements(statements_inside_scope, namespaces + [{'event': Name('event', get_mouse_event_object(x, y, button, event.value))}], [], when_statement_watchers + [{}])
                    del mouse_buttons[button]
            listener = mouse.Listener(on_click=listener_func)  # type: ignore
 
        case "mousedown":
            def listener_func(x: int, y: int, button: mouse.Button, pressed: bool):
                nonlocal namespaces, statements_inside_scope
                if pressed:
                    interpret_code_statements(statements_inside_scope, namespaces + [{'event': Name('event', get_mouse_event_object(x, y, button, event.value))}], [], when_statement_watchers + [{}])
            listener = mouse.Listener(on_click=listener_func)  # type: ignore

        case "mouseup":
            def listener_func(x: int, y: int, button: mouse.Button, pressed: bool):
                nonlocal namespaces, statements_inside_scope
                if not pressed:
                    interpret_code_statements(statements_inside_scope, namespaces + [{'event': Name('event', get_mouse_event_object(x, y, button, event.value))}], [], when_statement_watchers + [{}])
            listener = mouse.Listener(on_click=listener_func)  # type: ignore

        case "keyclick":
            keys = set()
            def on_press(key: Optional[Union[keyboard.Key, keyboard.KeyCode]]):
                nonlocal namespaces, statements_inside_scope
                keys.add(key)
            def on_release(key: Optional[Union[keyboard.Key, keyboard.KeyCode]]):
                nonlocal namespaces, statements_inside_scope
                if key in keys:
                    interpret_code_statements(statements_inside_scope, namespaces + [{'event': Name('event', get_keyboard_event_object(key, event.value))}], [], when_statement_watchers + [{}])
                keys.discard(key)
            listener = keyboard.Listener(on_press=on_press, on_release=on_release)  # type: ignore

        case "keydown":
            def on_press(key: Optional[Union[keyboard.Key, keyboard.KeyCode]]):
                nonlocal namespaces, statements_inside_scope
                interpret_code_statements(statements_inside_scope, namespaces + [{'event': Name('event', get_keyboard_event_object(key, event.value))}], [], when_statement_watchers + [{}])
            listener = keyboard.Listener(on_press=on_press)  # type: ignore

        case "keyup":
            def listener_func(x: int, y: int, button: mouse.Button, pressed: bool):
                nonlocal namespaces, statements_inside_scope
                if not pressed:
                    interpret_code_statements(statements_inside_scope, namespaces + [{'event': Name('event', get_mouse_event_object(x, y, button, event.value))}], [], when_statement_watchers + [{}])
            listener = keyboard.Listener(on_click=listener_func)  # type: ignore

        case _:
            raise_error_at_line(filename, code, current_line, f"Invalid event for the \"after\" statement: \"{db_to_string(event)}\"")

    listener.start()

def gather_names_or_values(expr: ExpressionTreeNode) -> set[Token]:
    names: set[Token] = set()
    match expr:
        case FunctionNode():
            for arg in expr.args:
                names |= gather_names_or_values(arg)
        case ListNode():
            for val in expr.values:
                names |= gather_names_or_values(val)
        case ExpressionNode():
            names |= gather_names_or_values(expr.right) | gather_names_or_values(expr.left)
        case IndexNode():
            names |= gather_names_or_values(expr.index) | gather_names_or_values(expr.value)
        case SingleOperatorNode():
            names |= gather_names_or_values(expr.expression)
        case ValueNode():
            names.add(expr.name_or_value)
    return names

def register_when_statement(condition: Union[list[Token], ExpressionTreeNode], statements_inside_scope: list[tuple[CodeStatement, ...]], namespaces: list[Namespace], async_statements: AsyncStatements, when_statement_watchers: WhenStatementWatchers):

    # if it is a variable, store it as the address to that variable.
    # if the internal value is a list, store it as an address to that mutable type.
    built_condition = get_built_expression(condition)
    gathered_names = gather_names_or_values(built_condition)
    caller_names = [n for name in gathered_names if (n := '.'.join(name.value.split('.')[:-1]))]
    dict_keys = [id(v) if isinstance(v := get_name_from_namespaces(name.value, namespaces), Variable) else name.value for name in gathered_names]\
                + [id(v.value) for name in gathered_names if (v := get_name_from_namespaces(name.value, namespaces)) is not None and isinstance(v.value, DreamberdMutable)]\
                + [id(v) for name in caller_names if isinstance(v := get_name_from_namespaces(name, namespaces), Variable)]\
                + [id(v.value) for name in caller_names if (v := get_name_from_namespaces(name, namespaces)) is not None and isinstance(v.value, DreamberdMutable)]\

    # the last comprehension watches callers of things (like list in list.length), and requires some implementation in the evaluate_expression function 
    # so that the caller of a function is also observed for it being called
    
    # register for future whens
    for name in dict_keys:
        if name not in when_statement_watchers[-1]:
            when_statement_watchers[-1][name] = []
        when_statement_watchers[-1][name].append((built_condition, statements_inside_scope))

    # check the condition now
    condition_value = evaluate_expression(built_condition, namespaces, async_statements, when_statement_watchers)
    execute_conditional(condition_value, statements_inside_scope, namespaces, when_statement_watchers)
    
def interpret_statement(statement: CodeStatement, namespaces: list[Namespace], async_statements: AsyncStatements, when_statement_watchers: WhenStatementWatchers) -> Optional[Value]:

    # build a list of expressions that are modified to allow for the next keyword
    expressions_to_check: list[Union[list[Token], ExpressionTreeNode]] = []
    match statement:
        case VariableAssignment(): expressions_to_check = [statement.expression] + statement.indexes
        case VariableDeclaration() | Conditional() | AfterStatement() | ExpressionStatement(): expressions_to_check = [statement.expression]
    all_normal_nexts: set[tuple[str, int]] = set()
    all_async_nexts: set[str] = set()
    next_filtered_exprs: list[ExpressionTreeNode] = []
    for expr in expressions_to_check:
        built_expr = get_built_expression(expr)
        new_expr, normal_nexts, async_nexts = handle_next_expressions(built_expr, namespaces)
        all_normal_nexts |= normal_nexts 
        all_async_nexts |= async_nexts
        next_filtered_exprs.append(new_expr)
    prev_namespace: Namespace = {}
    all_nexts = {s[0] for s in all_normal_nexts} | all_async_nexts
    for expr in next_filtered_exprs:
        prev_namespace |= save_previous_values_next_expr(expr, all_nexts, namespaces)

    # we replace the expression stored in each statement with the newly built one  -- TODO: consider changing this as to not override expressions
    match statement:
        case VariableAssignment():
            statement.expression = next_filtered_exprs[0]
            statement.indexes = next_filtered_exprs[1:]
        case VariableDeclaration() | Conditional() | AfterStatement() | ExpressionStatement():
            statement.expression = next_filtered_exprs[0]
    if all_normal_nexts:
        match statement:  # this is here to make sure everything is going somewhat smoothly
            case VariableAssignment() | VariableDeclaration() | Conditional() | AfterStatement() | ExpressionStatement(): pass
            case _:
                raise_error_at_line(filename, code, current_line, "Something went wrong. It's not your fault, it's mine.")
        adjust_for_normal_nexts(statement, all_async_nexts, all_normal_nexts, None, namespaces, prev_namespace)
        return   # WE ARE EXITING HERE!!!!!!!!!!!!!!!!!!!!!!!!!!

    if all_async_nexts:  # temporarily put prev_next_valeus (blah blah blah) into the namespace along with those that are generated from await next
        namespaces[-1] |= (prev_namespace := prev_namespace | wait_for_async_nexts(all_async_nexts, namespaces))

    # finally actually execute the damn thing
    retval = None
    match statement:
        case VariableAssignment():
            assign_variable(
                statement, [evaluate_expression(expr, namespaces, async_statements, when_statement_watchers) for expr in statement.indexes], 
                evaluate_expression(statement.expression, namespaces, async_statements, when_statement_watchers),
                namespaces, async_statements, when_statement_watchers
            )

        case FunctionDefinition():
            namespaces[-1][statement.name.value] = Name(statement.name.value, DreamberdFunction(
                args = [arg.value for arg in statement.args],
                code = statement.code,
                is_async = statement.is_async
            ))

        case DeleteStatement(): 
            val, ns = get_name_and_namespace_from_namespaces(statement.name.value, namespaces)
            if val and ns:
                del ns[statement.name.value]
            deleted_values.add(determine_non_name_value(statement.name))

        case ExpressionStatement():
            val = evaluate_expression(statement.expression, namespaces, async_statements, when_statement_watchers)
            print_expression_debug(statement.debug, statement.expression, val, namespaces)

        case Conditional():
            condition = db_to_boolean(evaluate_expression(statement.expression, namespaces, async_statements, when_statement_watchers))
            retval = execute_conditional(condition, statement.code, namespaces, when_statement_watchers)

        case AfterStatement():  # create event listener
            event = evaluate_expression(statement.expression, namespaces, async_statements, when_statement_watchers)
            execute_after_statement(event, statement.code, namespaces, when_statement_watchers)

        case WhenStatement():  # create variable listener  # one more left !!! :D
            register_when_statement(statement.expression, statement.code, namespaces, async_statements, when_statement_watchers)
            
        case VariableDeclaration():
            declare_new_variable(statement, evaluate_expression(
                statement.expression, namespaces, async_statements, when_statement_watchers
            ), namespaces, async_statements, when_statement_watchers)

        case ImportStatement():
            for name in statement.names:
                if not (v := importable_names.get(name.value)):
                    raise_error_at_token(filename, code, f"Name {name.value} could not be imported.", name)
                namespaces[-1][name.value] = Name(name.value, v)

        case ExportStatement():
            for name in statement.names:
                if not (v := get_name_from_namespaces(name.value, namespaces)):
                    raise_error_at_token(filename, code, "Tried to export name that is not a name or variable.", name)
                exported_names.append((statement.target_file.value, name.value, v.value))

        case ClassDeclaration(): 

            class_namespace: Namespace = {}
            fill_class_namespace(statement.code, namespaces, class_namespace, async_statements)

            # create a builtin function that is a closure returning an object of that class
            instance_made = False
            class_name = statement.name

            def class_object_closure(*args: Value) -> DreamberdObject:
                nonlocal instance_made, class_namespace, class_name
                if instance_made:
                    raise_error_at_line(filename, code, current_line, f"Already made instance of the class \"{class_name}\".")
                instance_made = True 
                obj = DreamberdObject(class_name.value, class_namespace)
                if constructor := class_namespace.get(class_name.value):
                    args = [obj] + list(args)  # type: ignore
                    if not isinstance(func := constructor.value, DreamberdFunction):
                        raise_error_at_line(filename, code, current_line, "Cannot create class variable with the same name as the class.")
                    if len(func.args) > len(args):
                        raise_error_at_line(filename, code, current_line, f"Expected more arguments for function call with {len(func.args)} argument{'s' if len(func.args) != 1 else ''}.")
                    new_namespace: Namespace = {name: Name(name, arg) for name, arg in zip(func.args, args)}
                    interpret_code_statements(func.code, namespaces + [new_namespace], [], when_statement_watchers + [{}])
                    del class_namespace[class_name.value]  # remove the constructor
                return obj

            namespaces[-1][statement.name.value] = Name(statement.name.value, BuiltinFunction(-1, class_object_closure))

    clear_temp_namespace(namespaces, prev_namespace)
    return retval

def fill_class_namespace(statements: list[tuple[CodeStatement, ...]], namespaces: list[Namespace], class_namespace: Namespace, async_statements: AsyncStatements) -> None:
    for possible_statements in statements:
        statement = determine_statement_type(possible_statements, namespaces)
        match statement:
            case FunctionDefinition():
                if "this" in statement.args:
                    raise_error_at_line(filename, code, current_line, "\"this\" keyword not allowed in class function declaration arguments.")
                class_namespace[statement.name.value] = Name(statement.name.value, DreamberdFunction(
                    args = ["this"] + [arg.value for arg in statement.args],    # i need to somehow make the this keyword actually do something
                    code = statement.code,
                    is_async = statement.is_async
                ))
            case VariableDeclaration(): 

                # why the frick are my function calls so long i really need some globals  
                var_expr = evaluate_expression(statement.expression, namespaces, async_statements, [{}]) 
                declare_new_variable(statement, var_expr, namespaces + [class_namespace], async_statements, [{}])  # don't want anything happening here, it's a different name
            case _:
                raise_error_at_line(filename, code, current_line, f"Unexpected statement of type {type(statement).__name__} in class declaration.")

# cleanup out of date variables
def decrement_variable_lifetimes(namespaces: list[Namespace]) -> None:
    for ns in namespaces: 
        remove_vars = []
        for name, v in ns.items():
            if not isinstance(v, Variable):
                continue 
            for l in v.lifetimes:
                l.lines_left -= 1 
            v.clear_outdated_lifetimes()
            if not v.lifetimes:
                remove_vars.append(name)
        for name in remove_vars:
            del ns[name]  # simply remove the variable from the name as it will be unable to provide a value

# change the current line number to some token in the code
def edit_current_line_number(statement: CodeStatement) -> None:
    global current_line
    match statement:
        case CodeStatementKeywordable(): 
            current_line = statement.keyword.line
        case FunctionDefinition():
            current_line = statement.keywords[0].line 
        case VariableDeclaration():
            current_line = statement.modifiers[0].line 
        case VariableAssignment():
            current_line = statement.name.line
        case ReturnStatement(): 
            if statement.keyword:
                current_line = statement.keyword.line
        case ExpressionStatement():
            if t := get_expr_first_token(get_built_expression(statement.expression)):
                current_line = t.line

# if a return statement is found, this will return the expression evaluated at the return. otherwise, it will return None
# this is done to allow this function to be called when evaluating dreamberd functions
def interpret_code_statements(statements: list[tuple[CodeStatement, ...]], namespaces: list[Namespace], async_statements: AsyncStatements, when_statement_watchers: WhenStatementWatchers) -> Optional[Value]:

    curr, direction = 0, 1
    while 0 <= curr < len(statements): 
        decrement_variable_lifetimes(namespaces)
        statement = determine_statement_type(statements[curr], namespaces)

        # if no statement found -i.e. the user did something wrong
        if statement is None:
            raise_error_at_line(filename, code, current_line, "Error parsing statement. Try again.")
        edit_current_line_number(statement)
        if isinstance(statement, ReturnStatement):  # if working with a return statement, either return a promise or a value
            expr, normal_nexts, async_nexts = handle_next_expressions(get_built_expression(statement.expression), namespaces)
            prev_namespaces = save_previous_values_next_expr(expr, async_nexts | {s[0] for s in normal_nexts}, namespaces)
            if normal_nexts:
                promise = DreamberdPromise(None)
                adjust_for_normal_nexts(statement, async_nexts, normal_nexts, promise, namespaces, prev_namespaces)
                return promise
            elif async_nexts:
                namespaces[-1] |= (prev_namespaces := prev_namespaces | wait_for_async_nexts(async_nexts, namespaces))
            retval = evaluate_expression(expr, namespaces, async_statements, when_statement_watchers)
            clear_temp_namespace(namespaces, prev_namespaces)
            return retval
        elif isinstance(statement, ReverseStatement):
            direction = -direction
        
        # otherwise interpret the other statement, if there is a return value then return
        elif retval := interpret_statement(statement, namespaces, async_statements, when_statement_watchers):
            return retval
        curr += direction

        # emulate taking turns running all the "async" functions
        for i in range(len(async_statements)):
            async_st, async_ns, line_num, async_direction = async_statements[i]
            if not 0 <= line_num < len(async_st):
                continue

            # this uesd to say async_st.pop(0) and i was genuinely wondering why it was changing, i'm so dumb
            statement = determine_statement_type(async_st[line_num], async_ns)
            if statement is None:
                raise_error_at_line(filename, code, current_line, "Error parsing statement. Try again.")
            elif isinstance(statement, ReturnStatement):
                raise_error_at_line(filename, code, current_line, "Function executed asynchronously cannot have a return statement.")
            elif isinstance(statement, ReverseStatement):
                async_direction = -async_direction
            async_statements[i] = (async_st, async_ns, line_num + async_direction, async_direction)  # messy but works
            interpret_statement(statement, async_ns, async_statements, when_statement_watchers)
    
# btw, reason async_statements and when_statements cannot be global is because they change depending on scope,
# due to (possibly bad) design decisions, the name_watchers does not do this... :D
def load_globals(_filename: str, _code: str, _name_watchers: NameWatchers, _deleted_values: set[Value], _exported_names: list[tuple[str, str, Value]], _importable_names: dict[str, Value]):
    global filename, code, name_watchers, deleted_values, current_line, exported_names, importable_names  # screw bad practice, not like anyone's using this anyways
    filename = _filename 
    code = _code
    name_watchers = _name_watchers 
    deleted_values = _deleted_values
    exported_names = _exported_names
    importable_names = _importable_names
    current_line = 1
    
def interpret_code_statements_main_wrapper(statements: list[tuple[CodeStatement, ...]], namespaces: list[Namespace], async_statements: AsyncStatements, when_statement_watchers: WhenStatementWatchers):
    try:
        interpret_code_statements(statements, namespaces, async_statements, when_statement_watchers)
    except NonFormattedError as e:
        raise_error_at_line(filename, code, current_line, str(e))



# basic things like the token class, keywords, etc.     
from __future__ import annotations

from enum import Enum
from typing import NoReturn
from dataclasses import dataclass, field

ALPH_NUMS = set('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_.')

class NonFormattedError(Exception): pass

class InterpretationError(Exception):
    _           :(                                                                                              str)  # this is why i am the greatest programmer to ever live

def debug_print(filename: str, code: str, message: str, token: Token) -> None:
    if not code:  # adjust for repl-called code
        print(f'\n\033[33m{message}\033[39\n', sep="")
        return
    line = token.line
    num_carrots, num_spaces = len(token.value), token.col - len(token.value) + 1
    debug_string = f"\033[33m{filename}, line {line}\033[39m\n\n" + \
                   f"  {code.split(chr(10))[line - 1]}\n" + \
                   f" {num_spaces * ' '}{num_carrots * '^'}\n" + \
                   f"\033[33m{message}\033[39m"
    print('\n', debug_string, '\n', sep="")

def debug_print_no_token(filename: str, message: str) -> None:
    debug_string = f"\033[33m{filename}\033[39m\n\n" + \
                   f"\033[33m{message}\033[39m"
    print('\n', debug_string, '\n', sep="")

def raise_error_at_token(filename: str, code: str, message: str, token: Token) -> NoReturn:
    if not code:  # adjust for repl-called code
        raise InterpretationError(f"\n\033[31m{message}\033[39m\n")
    line = token.line
    num_carrots, num_spaces = len(token.value), token.col - len(token.value) + 1
    error_string = f"\033[33m{filename}, line {line}\033[39m\n\n" + \
                   f"  {code.split(chr(10))[line - 1]}\n" + \
                   f" {num_spaces * ' '}{num_carrots * '^'}\n" + \
                   f"\033[31m{message}\033[39m"
    raise InterpretationError(error_string)

def raise_error_at_line(filename: str, code: str, line: int, message: str) -> NoReturn:
    if not code:  # adjust for repl-called code
        raise InterpretationError(f"\n\033[31m{message}\033[39m\n")
    error_string = f"\033[33m{filename}, line {line}\033[39m\n\n" + \
                   f"  {code.split(chr(10))[line - 1]}\n\n" + \
                   f"\033[31m{message}\033[39m"
    raise InterpretationError(error_string)

class TokenType(Enum): 
    R_CURLY = '}'
    L_CURLY = '{'
    R_SQUARE = ']'
    L_SQUARE = '['

    DOT = '.'
    ADD = '+'
    INCREMENT = '++'
    DECREMENT = '--'
    EQUAL = '='
    DIVIDE = '/'
    MULTIPLY = '*'
    SUBTRACT = '-'

    COMMA = ','
    COLON = ':'
    SEMICOLON = ';'
    BANG = '!'
    QUESTION = '?'
    CARROT = '^'
    FUNC_POINT = '=>'

    LESS_THAN = '<'
    GREATER_THAN = '>'
    LESS_EQUAL = '<='
    GREATER_EQUAL = '>='
    NOT_EQUAL = ';='  #!@#!@#!@#
    PIPE = '|'
    AND = '&'

    WHITESPACE = '       '
    NAME = 'abcaosdijawef'  # i'm losing my mind
    STRING = "'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''"  # iasddddakdjhnakjsndkjsbndfkijewbgf

    NEWLINE = '\n'
    SINGLE_QUOTE = "'"  # this is ugly as hell
    DOUBLE_QUOTE = '"'

class OperatorType(Enum):
    ADD = '+'
    SUB = '-'
    MUL = '*'
    DIV = '/'
    EXP = '^'
    GT  = '>'
    GE  = '>='
    LT  = '<'
    LE  = '<='
    OR  = '|'
    AND = '&'
    COM = ','  # this is just here to seperate variables in a function 
    E   = '='
    EE  = '=='
    EEE = '==='
    EEEE= '===='
    NE  = ';='
    NEE = ';=='
    NEEE= ';==='

STR_TO_OPERATOR = {op.value: op for op in OperatorType}

# why do i even need a damn class for this 
# 3 weeks later, i am very glad i made a class for this
@dataclass(unsafe_hash=True)
class Token():

    type: TokenType
    value: str 
    line: int = field(hash=False)
    col: int = field(hash=False)

    def __repr__(self) -> str:
        return f"Token({self.type}, {repr(self.value)})"



from __future__ import annotations
from abc import ABCMeta, abstractmethod
from typing import Optional

from dreamberd.base import STR_TO_OPERATOR, NonFormattedError, Token, TokenType, OperatorType, InterpretationError, raise_error_at_token

class ExpressionTreeNode(metaclass=ABCMeta):
    @abstractmethod
    def to_string(self, tabs: int = 0) -> str: pass

# things like the not operator
class SingleOperatorNode(ExpressionTreeNode):
    def __init__(self, expression: ExpressionTreeNode, operator: Token) -> None:
        self.expression = expression
        self.operator = operator
    def to_string(self, tabs: int = 0) -> str:
        return f"{'  ' * tabs}Single Operator: \n" + \
               f"{'  ' * (tabs + 1)}Operator: {self.operator.value}\n" + \
               f"{'  ' * (tabs + 1)}Expression: \n" + \
               f"{self.expression.to_string(tabs + 2)}"
        
class ListNode(ExpressionTreeNode):
    def __init__(self, values: list[ExpressionTreeNode]):
        self.values = values
    def to_string(self, tabs: int = 0) -> str:
        return f"{'  ' * tabs}List: \n" + \
               f"{'  ' * (tabs + 1)}Values: \n" + \
               "\n".join([f"{v.to_string(tabs + 2)}" for v in self.values])

class ExpressionNode(ExpressionTreeNode):
    def __init__(self, left: ExpressionTreeNode, right: ExpressionTreeNode, operator: OperatorType, operator_token: Token):
        self.left = left 
        self.right = right
        self.operator = operator
        self.operator_token = operator_token
    def to_string(self, tabs: int = 0) -> str:
        return f"{'  ' * tabs}Expression: \n" + \
               f"{'  ' * (tabs + 1)}Operator: {self.operator}\n" + \
               f"{'  ' * (tabs + 1)}Left: \n" + \
               f"{self.left.to_string(tabs + 2)}\n" + \
               f"{'  ' * (tabs + 1)}Right: \n" + \
               f"{self.right.to_string(tabs + 2)}"

class FunctionNode(ExpressionTreeNode):
    def __init__(self, name: Token, args: list[ExpressionTreeNode]):
        self.name = name 
        self.args = args
    def to_string(self, tabs: int = 0) -> str:
        return f"{'  ' * tabs}Function: \n" + \
               f"{'  ' * (tabs + 1)}Name: {self.name}\n" + \
               f"{'  ' * (tabs + 1)}Arguments: \n" + \
               "\n".join([f"{arg.to_string(tabs + 2)}" for arg in self.args]) 

class IndexNode(ExpressionTreeNode):
    def __init__(self, value: ExpressionTreeNode, index: ExpressionTreeNode):
        self.value = value 
        self.index = index
    def to_string(self, tabs: int = 0) -> str:
        return f"{'  ' * tabs}Index: \n" + \
               f"{'  ' * (tabs + 1)}Of: \n" + \
               f"{self.value.to_string(tabs + 2)}\n" + \
               f"{'  ' * (tabs + 1)}At: \n" + \
               f"{self.index.to_string(tabs + 2)}"

class ValueNode(ExpressionTreeNode):
    def __init__(self, name_or_value: Token): 
        self.name_or_value = name_or_value
    def to_string(self, tabs: int = 0) -> str:
        return f"{'  ' * tabs}Value: {self.name_or_value}"

def get_expr_first_token(expr: ExpressionTreeNode) -> Optional[Token]:
    match expr:
        case SingleOperatorNode(): return expr.operator 
        case ExpressionNode(): return get_expr_first_token(expr.left) or get_expr_first_token(expr.right)
        case FunctionNode(): return expr.name 
        case ListNode(): return None if not expr.values else get_expr_first_token(expr.values[0])
        case ValueNode(): return expr.name_or_value 
        case IndexNode(): return get_expr_first_token(expr.value) or get_expr_first_token(expr.index)

def build_expression_tree(filename: str, tokens: list[Token], code: str) -> ExpressionTreeNode:
    """ 
    This language has significant whitespace, so the biggest split happens where there is most space
     - func a, b  +  c becomes func(a, b) + c but func a, b+c  becomes func(a, b + c) 
     - a + func   b  ,  c + d is not legal because it translates to (a + func)(b, c + d)
     - 2 * 1+3 becomes 2 * (1 + 3)
    """

    if not tokens:
        raise NonFormattedError("Something went wrong, I don't know what so figure it out :)")

    # tabs at the beginning or end do not matter
    for token in tokens[1:-1]:
        if token.type == TokenType.WHITESPACE and '\t' in token.value:
            raise_error_at_token(filename, code, "Tabs are not allowed in expressions.", token)
        elif token.type == TokenType.NEWLINE:
            raise_error_at_token(filename, code, "Due to the laws of significant whitespace, no newline characters are permitted in expressions. If your code is so long that it needs newlines, consider rewriting it :)", token)
    
    # create a new list consisting and tokens and a brand new type: the list 
    tokens_without_whitespace = [token for token in tokens if token.type != TokenType.WHITESPACE]
    if len(tokens_without_whitespace) == 2 and tokens_without_whitespace[0].type == TokenType.L_SQUARE and tokens_without_whitespace[1].type == TokenType.R_SQUARE:
        return ListNode([])  # easy way out xD
    starts_with_whitespace = tokens[0].type == TokenType.WHITESPACE
    ends_with_whitespace = tokens[-1].type == TokenType.WHITESPACE

    # transform a list of tokens to include operators 
    # find the operator with the maximum whitespace between it and other things
    updated_list = [STR_TO_OPERATOR.get(token.value, token) for token in tokens]
    max_width, max_index = -1, -1
    bracket_layers = 0
    for i in range(len(updated_list)):
        if tokens[i].type == TokenType.L_SQUARE:
            bracket_layers += 1
        elif tokens[i].type == TokenType.R_SQUARE:
            bracket_layers -= 1
        if isinstance(updated_list[i], OperatorType) and bracket_layers == 0:
            try:

                # first check for negative sign -- this is horrible :D 
                if i == 0 or tokens[i - 1].type == TokenType.WHITESPACE and \
                   tokens[i + 1].type != TokenType.WHITESPACE and updated_list[i] == OperatorType.SUB:  # l_len greater than zero no matter what fr
                    continue

                # make sure whitespace is equal and then determine maxes
                l_len, r_len = 0, 0
                if tokens[i - 1].type == TokenType.WHITESPACE:
                    l_len = len(tokens[i - 1].value)
                if tokens[i + 1].type == TokenType.WHITESPACE:
                    r_len = len(tokens[i + 1].value)
                if l_len != r_len and updated_list[i] != OperatorType.COM:
                    raise_error_at_token(filename, code, "Whitespace must be equal on either side of an operator.", tokens[i])
                if r_len >= max_width:
                    max_width = r_len
                    max_index = i

            except IndexError:
                raise_error_at_token(filename, code, "Operator cannot be at the end of an expression.", tokens[i])

    # detecting single argument function
    # this doesn't seem to adhere to my standards 100%, so its not a bug, its a feature
    starts_with_operator = int(tokens_without_whitespace[0].type in {TokenType.SEMICOLON, TokenType.SUBTRACT})
    first_name_index = int(starts_with_whitespace) + int(starts_with_operator)
    if len(tokens) >= 3 + first_name_index and \
       tokens[first_name_index].type == TokenType.NAME and \
       tokens[first_name_index + 1].type == TokenType.WHITESPACE and \
       tokens[first_name_index + 2].type in [TokenType.NAME, TokenType.L_SQUARE, TokenType.STRING, TokenType.SUBTRACT, TokenType.SEMICOLON] and \
       len(tokens[first_name_index + 1].value) > max_width:
        function_node = FunctionNode(tokens[first_name_index],
                                     [build_expression_tree(filename, tokens[first_name_index + 1:], code)])
        if starts_with_operator:
            return SingleOperatorNode(function_node, tokens_without_whitespace[0])
        return function_node

    # check if there is an operator at the beginning of the thing
    if starts_with_operator and (max_index == -1 or (t := tokens[int(starts_with_whitespace) + 1]).type == TokenType.WHITESPACE and \
        len(t.value) > max_width or updated_list[max_index] == OperatorType.COM):
        return SingleOperatorNode(build_expression_tree(filename, tokens[int(starts_with_whitespace) + 1:], code), tokens_without_whitespace[0])

    # value, like a list, name, or anything else
    if max_index == -1:

        # just making sure the input is correct
        try:
            name_or_value = tokens_without_whitespace[0]
            if name_or_value.type not in [TokenType.NAME, TokenType.L_SQUARE, TokenType.STRING]:
                raise_error_at_token(filename, code, "Expected name or value.", tokens_without_whitespace[0])
        except IndexError:
            raise_error_at_token(filename, code, "Expected name or value.", tokens_without_whitespace[0])

        # this is a list :)
        if name_or_value.type == TokenType.L_SQUARE:
            bracket_layers = 1
            for i, token in enumerate(tokens_without_whitespace[1:], start=1):
                if token.type == TokenType.L_SQUARE:
                    bracket_layers += 1 
                elif token.type == TokenType.R_SQUARE:
                    bracket_layers -= 1

                # this means the closing happen, signifying the end of the list
                if bracket_layers == 0:
                    if i == len(tokens_without_whitespace) - 1:
    
                        # let's find the most significant comma, and split by that
                        # if there's a function in the middle of the list, too bad :)
                        # [func a, b]  == [func(a), b] and also [func(a, b)]  # literally how do i tell them apart

                        # need to consider the width of whitespace from either side fr
                        l_width = len(token.value) if (token := tokens[int(starts_with_whitespace) + 1]).type == TokenType.WHITESPACE else 0
                        r_width = len(token.value) if (token := tokens[len(tokens) - int(ends_with_whitespace) - 2]).type == TokenType.WHITESPACE else 0
                        if l_width != r_width:
                            raise_error_at_token(filename, code, "Whitespace between either bracket of a list must be equal in length.",
                                                 tokens[len(tokens) - int(ends_with_whitespace) - 2])

                        # now go through all the commas and check if the whitespace is significant
                        all_commas = []
                        bracket_layers = 0  # yes i'm setting this damn thing twice 
                        for i, (token, tok_or_op) in enumerate(zip(tokens[:-1], updated_list)):  # stop here to avoid angry errors
                            if token.type == TokenType.L_SQUARE:
                                bracket_layers += 1 
                            elif token.type == TokenType.R_SQUARE:
                                bracket_layers -= 1
                            if tok_or_op == OperatorType.COM and bracket_layers == 1 and (
                                l_width == 0 or l_width == len(tokens[i + 1].value) and tokens[i + 1].type == TokenType.WHITESPACE
                            ):
                                all_commas.append(i)

                        # not single element
                        if all_commas:
                            return ListNode([
                                build_expression_tree(filename, t, code) for t in [
                                    tokens[comma_index + 1:next_index] for comma_index, next_index in 
                                    zip([int(starts_with_whitespace), *all_commas], 
                                        [*all_commas, len(tokens) - 1 - int(ends_with_whitespace)])  # adjusting here in order to avoid the bracket tokens
                                ]
                            ])
                        
                        # single element :)
                        return ListNode([build_expression_tree(filename, tokens[int(starts_with_whitespace) + 1 : len(tokens) - int(ends_with_whitespace) - 1], code)])
                    break

        # now we need to handle indexes
        # let's go from the back of the list and find the first fully closing sequence 
        # i am sure that this guarantees there is an index (i think)
        if tokens_without_whitespace[-1].type == TokenType.R_SQUARE:
            bracket_layers = -1
            end_index = len(tokens) - int(ends_with_whitespace) - 1
            for i, token in reversed(list(enumerate(tokens[:end_index]))):  # i don't like this one bit  :(
                if token.type == TokenType.L_SQUARE:
                    bracket_layers += 1 
                elif token.type == TokenType.R_SQUARE:
                    bracket_layers -= 1

                # first index!!!!!!!!!!!!!!!!!!!!
                if bracket_layers == 0:
                    return IndexNode(build_expression_tree(filename, tokens[int(starts_with_whitespace) : i], code),
                                     build_expression_tree(filename, tokens[i + 1 : end_index], code))
                    
        # finally end this vicious cycle
        return ValueNode(name_or_value)
        
    # max_index is the token with the maximum surrouding whitespace 
    if updated_list[max_index] == OperatorType.COM:  
        # this means it is a function
        # we need to find every other comma as they become the arguments of the function
        # additionally, there needs to be a spacing of equal length between the name of the function and the next argument
        
        if tokens_without_whitespace[0].type != TokenType.NAME or \
           tokens_without_whitespace[1].type not in [TokenType.NAME, TokenType.L_SQUARE, TokenType.STRING]:
            raise_error_at_token(filename, code, "Expected function call. This is likely an issue of whitespace, as DreamBerd replaces parentheses with spaces and has significant whitespace.", tokens_without_whitespace[0])
        
        all_commas = []
        for i in range(len(updated_list)):
            if updated_list[i].value == ',':  # okay this is weird because enums have .value and tokens have .value
                if max_width == 0 or (tokens[i + 1].type == TokenType.WHITESPACE and len(tokens[i + 1].value)) == max_width:
                    all_commas.append(i)
        
        # i have no idea what the hell im doin
        return FunctionNode(tokens_without_whitespace[0], [
            build_expression_tree(filename, t, code) for t in [
                tokens[comma_index + 1:next_index] for comma_index, next_index in 
                zip([int(starts_with_whitespace), *all_commas], [*all_commas, len(tokens)])
            ]
        ])

    else: 
        operator = updated_list[max_index]
        if not isinstance(operator, OperatorType): 
            raise_error_at_token(filename, code, "Something went wrong. My bad.", tokens[max_index])
        return ExpressionNode(
            build_expression_tree(filename, tokens[:max_index], code), 
            build_expression_tree(filename, tokens[max_index + 1:], code), 
            operator=operator,
            operator_token=tokens[max_index]
        )



import re 
import sys
from argparse import ArgumentParser
from time import sleep
from typing import Optional, Union
from dreamberd.base import InterpretationError, NonFormattedError, Token, TokenType

from dreamberd.builtin import KEYWORDS, Name, Value, Variable
from dreamberd.processor.lexer import tokenize
from dreamberd.processor.syntax_tree import generate_syntax_tree
from dreamberd.interpreter import interpret_code_statements, interpret_code_statements_main_wrapper, load_global_dreamberd_variables, load_globals, load_public_global_variables

__all__ = ['run_repl', 'run_file']

__REPL_FILENAME = "__repl__"
sys.setrecursionlimit(100000)

def __get_next_repl_input(closed_scope_layers: int = 0) -> tuple[str, list[Token]]:
    print("   " * closed_scope_layers, '\033[33m>\033[39m ', end="", sep="")
    code = input()
    tokens = tokenize(__REPL_FILENAME, code)
    start_closed_scope_layers = closed_scope_layers
    for t in tokens:
        if t.type == TokenType.L_CURLY:
            closed_scope_layers += 1
        elif t.type == TokenType.R_CURLY:
            closed_scope_layers -= 1
    if closed_scope_layers < 0:
        raise InterpretationError("Too many closed braces with not enough open braces.")
    kickback_layers, i = 0, 0
    trimmed_code = code.strip()
    while i < len(trimmed_code) and trimmed_code[i] == '}':
        kickback_layers += 1
        i += 1
    if kickback_layers:
        print("\033[1A\r", "   " * (start_closed_scope_layers - kickback_layers), '\033[33m>\033[39m ', code, "   " * (start_closed_scope_layers), sep="", flush=True)
    if closed_scope_layers:
        new_code, new_tokens = __get_next_repl_input(closed_scope_layers) 
        code += new_code 
        tokens += new_tokens
    match l := [t for t in tokens if not t.type == TokenType.WHITESPACE]:
        case [Token(type=TokenType.NAME)]:
            tokens = [Token(TokenType.NAME, "print", -1, -1), Token(TokenType.WHITESPACE, " ", -1, -1), l[0], Token(TokenType.BANG, '!', -1, -1)]
    return code, tokens

def run_repl() -> None:
    namespaces: list[dict[str, Union[Variable, Name]]] = [KEYWORDS.copy()]  # type: ignore
    load_globals(__REPL_FILENAME, "", {}, set(), [])
    load_global_dreamberd_variables(namespaces)
    async_statements = []
    when_statement_watchers = [{}]
    while True: 
        try:
            code, tokens = __get_next_repl_input()
            statements = generate_syntax_tree(__REPL_FILENAME, tokens, code)
            interpret_code_statements(statements, namespaces, async_statements, when_statement_watchers)
        except InterpretationError as e:
            print(e)
        except NonFormattedError as e:
            print("\033[31m\n", e, "\n\033[39m", sep="")
            
def run_file(main_filename: str) -> None:  # idk what else to call this

    with open(main_filename, 'r') as f:
        code_lines = f.readlines()

    # split up into seperate 'files' by finding which lines start with multiple equal signs
    files: list[tuple[Optional[str], str]] = []
    if any(matches := [re.match(r"=====.*", l) for l in code_lines]):
        for i, match in reversed([*enumerate(matches)]):
            if match is None: continue 
            name = match.group().strip('=').strip() or None
            files.insert(0, (name, ''.join(code_lines[i+1:])))
            del code_lines[i:]
        files.insert(0, (None, ''.join(code_lines[0:])))
    else:
        files = [(None, ''.join(code_lines))]
    
    # execute code for each file
    importable_names: dict[str, dict[str, Value]] = {} 
    for filename, code in files:
        filename = filename or "__unnamed_file__"
        tokens = tokenize(filename, code)
        statements = generate_syntax_tree(filename, tokens, code)

        # load variables and run the code
        namespaces: list[dict[str, Union[Variable, Name]]] = [KEYWORDS.copy()]   # type: ignore
        exported_names: list[tuple[str, str, Value]] = []
        load_globals(filename, code, {}, set(), exported_names, importable_names.get(filename, {}))
        load_global_dreamberd_variables(namespaces)
        load_public_global_variables(namespaces)
        interpret_code_statements_main_wrapper(statements, namespaces, [], [{}])

        # take exported names and put them where they belong
        for target_filename, name, value in exported_names:
            if target_filename not in importable_names:
                importable_names[target_filename] = {}
            importable_names[target_filename][name] = value

    print("\033[33mCode has finished executing. Press ^C once or twice to stop waiting for when-statements and after-statements.\033[039m")
    try:
        while True:
            sleep(1)  # just waiting for any clicks, when statements, etc
    except KeyboardInterrupt:
        exit()  # quit silently

def parse_args():
    parser = ArgumentParser()
    parser.add_argument('file', help="the file containing your DreamBerd code", nargs='?', default='', type=str)
    parser.add_argument('-s', '--show-traceback', help="show the full Python trackback upon errors", action="store_true")
    return parser.parse_args()

def main():
    args = parse_args()
    if not args.show_traceback:
        sys.tracebacklimit = 0
    if not args.file:
        run_repl()
    else:
        run_file(args.file)

if __name__ == "__main__":
    main()



from abc import ABCMeta 
from typing import Optional, Union
from dataclasses import dataclass

from dreamberd.base import STR_TO_OPERATOR, Token, TokenType, raise_error_at_line, raise_error_at_token
from dreamberd.processor.expression_tree import ExpressionTreeNode

class CodeStatement():
    pass

class CodeStatementKeywordable(metaclass=ABCMeta):  # this is completely unnecessary lmao
    keyword: Token

class CodeStatementDebuggable(metaclass=ABCMeta):   # this is even more unnecessary
    debug: int

# name name argname, argname, argname => {  ...
# for single line arrows, this is:
# name name argname, => expression !?
@dataclass
class FunctionDefinition(CodeStatement):
    keywords: list[Token]
    name: Token
    args: list[Token]
    code: list[tuple[CodeStatement, ...]]
    is_async: bool

# name name {
@dataclass
class ClassDeclaration(CodeStatement, CodeStatementKeywordable):
    keyword: Token
    name: Token 
    code: list[tuple[CodeStatement, ...]]

# : ... can be ignored
# name name name: ... = something!?
@dataclass
class VariableDeclaration(CodeStatement, CodeStatementDebuggable):
    name: Token
    modifiers: list[Token]
    lifetime: Optional[str]
    expression: Union[list[Token], ExpressionTreeNode]
    debug: int
    confidence: int  ## amount of !!! after the decl for priority

# name []? = ...!
@dataclass 
class VariableAssignment(CodeStatement, CodeStatementDebuggable):
    name: Token
    expression: Union[list[Token], ExpressionTreeNode]
    debug: int
    indexes: Union[list[list[Token]], list[ExpressionTreeNode]]  # list[Token] here is an expression not evaled yet
    confidence: int 

# name expression { 
# since an expression can be a name:
#   name name {  
# can be both, and would have to be determined at runtime
@dataclass
class Conditional(CodeStatement, CodeStatementKeywordable):
    keyword: Token
    expression: Union[list[Token], ExpressionTreeNode]
    code: list[tuple[CodeStatement, ...]]

# name expression !?
@dataclass
class ReturnStatement(CodeStatement, CodeStatementDebuggable):
    keyword: Optional[Token]
    expression: Union[list[Token], ExpressionTreeNode]
    debug: int

# name name !?
@dataclass
class DeleteStatement(CodeStatement, CodeStatementKeywordable, CodeStatementDebuggable):
    keyword: Token
    name: Token
    debug: int

# name!
@dataclass
class ReverseStatement(CodeStatement, CodeStatementKeywordable, CodeStatementDebuggable):
    keyword: Token
    debug: int

# expression !?   < virtually indistinguishable from a return statement from a parsing perspective
@dataclass
class ExpressionStatement(CodeStatement, CodeStatementDebuggable):
    expression: Union[list[Token], ExpressionTreeNode]
    debug: int

# name name = expression { 
@dataclass
class WhenStatement(CodeStatement, CodeStatementKeywordable):
    keyword: Token
    expression: Union[list[Token], ExpressionTreeNode]
    code: list[tuple[CodeStatement, ...]]

# name "string" expression!
@dataclass 
class AfterStatement(CodeStatement, CodeStatementKeywordable):
    keyword: Token
    expression: Union[list[Token], ExpressionTreeNode]
    code: list[tuple[CodeStatement, ...]]

# name name (, name)* name string!
@dataclass 
class ExportStatement(CodeStatement, CodeStatementDebuggable):
    export_keyword: Token 
    names: list[Token]
    to_keyword: Token 
    target_file: Token
    debug: int

# name 
@dataclass 
class ImportStatement(CodeStatement, CodeStatementKeywordable, CodeStatementDebuggable):
    keyword: Token 
    names: list[Token]
    debug: int

# idea: create a class that evaluates at runtime what a statement is, so then execute it 
def split_into_statements(tokens: list[Token]) -> list[list[Token]]:
    statements = [[]]
    bracket_layers = 0
    for token in tokens:

        # check for expression-ending newlines
        if token.type == TokenType.WHITESPACE and not statements[-1]:  # don't care about whitespace at the beginning or end of an expression, idk
            continue
        if token.type == TokenType.NEWLINE and not statements[-1] and len(statements) > 1:
            statements[-2].append(token)
        else:
            statements[-1].append(token)

        # this is the start of a new scope, we don't care about those for rn in terms of starting a new statement
        if token.type == TokenType.L_CURLY:
            bracket_layers += 1 
        elif token.type == TokenType.R_CURLY:
            bracket_layers -= 1 

        if token.type in [TokenType.R_CURLY, TokenType.BANG, TokenType.QUESTION] and bracket_layers == 0:
            while statements[-1][-1].type == TokenType.NEWLINE:  # remove newlines at the end of a statement
                statements[-1].pop()
            statements.append([])
    
    # remove stray newlines cause they are annoying and shit, also remove empty tings
    final_statements = []
    for statement in statements:
        while statement and statement[-1].type in {TokenType.WHITESPACE, TokenType.NEWLINE}:  ## NOTE: END WILL NEVER BE WHITESPACE
            statement.pop()
        while statement and statement[0].type in {TokenType.WHITESPACE, TokenType.NEWLINE}:  ## NOTE: NOW START WILL NEVER BE WHITESPACE EITHER
            statement.pop(0)
        if not statement:
            continue
        final_statements.append(statement)

    return final_statements

def remove_type_hints(filename: str, code: str, statements: list[list[Token]]) -> list[list[Token]]:
    new_statements = []
    for tokens in statements:
        new_tokens = []
        adding_tokens = True 
        scope_layers, square_bracket_layers = 0, 0
        ref_square_bracket_layers = 0 
        curr = 0

        while curr < len(tokens):
            t = tokens[curr]

            # handle brackets
            if t.type == TokenType.L_CURLY: scope_layers += 1
            elif t.type == TokenType.R_CURLY: scope_layers -= 1
            if t.type == TokenType.L_SQUARE: square_bracket_layers += 1
            elif t.type == TokenType.R_SQUARE: square_bracket_layers -= 1


            # must be in the right place to consider removal
            if t.type == TokenType.COLON and scope_layers == 0:
                adding_tokens = False
                ref_square_bracket_layers = square_bracket_layers  # prob gonna be zero but idek imma just do this
            if not adding_tokens:

                # check if it is at an operator 
                if STR_TO_OPERATOR.get(t.value) and square_bracket_layers == ref_square_bracket_layers:
                    adding_tokens = True

                # adjust for Name<...> things, which also allows regex to pass too 
                elif t.type == TokenType.NAME and curr + 1 < len(tokens) and tokens[curr + 1].type == TokenType.LESS_THAN:
                    try:
                        while tokens[curr + 1].type != TokenType.GREATER_THAN:
                            curr += 1
                        curr += 1
                    except IndexError:
                        raise_error_at_token(filename, code, "Something went wrong parsing type hints (a.k.a. removing them).", t)

            if adding_tokens:  # cannot be elif because adding_tokens can be modified in the previous statement
                new_tokens.append(t)

            curr += 1
        
        new_statements.append(new_tokens)
    
    return new_statements

def assert_proper_indentation(filename: str, tokens: list[Token], code: str) -> None:
    looking_for_whitespace = False
    for t in tokens:
        if not looking_for_whitespace:
            if t.type == TokenType.NEWLINE:
                looking_for_whitespace = True
        else:
            if t.type == TokenType.WHITESPACE and len(t.value.replace('\t', '  ')) % 3:
                raise_error_at_token(filename, code, "Invalid indenting detected (must be a multiple of 3). Tabs count as 2 spaces.", t)
            looking_for_whitespace = False

def create_function_definition(filename: str, without_whitespace: list[Token], code: str, statements_inside_scope: list[tuple[CodeStatement, ...]]) -> tuple[CodeStatement, ...]:

    # parse this more to make sure it really really really can be a function
    names_in_row: list[Token] = []
    other_names: list[Token] = []
    looking_for_in_row = True
    for t in without_whitespace:
        if t.type == TokenType.FUNC_POINT:
            break
        if looking_for_in_row:
            if t.type != TokenType.NAME:
                looking_for_in_row = False
            elif t.value:
                names_in_row.append(t)
        else:
            if t.type == TokenType.NAME and t.value != "":
                other_names.append(t)
            elif t.type != TokenType.COMMA:
                raise_error_at_token(filename, code, "Invalid token in function declaration.", t)

    # too many or too little keywords for the function
    if not 2 <= len(names_in_row) <= 4:
        raise_error_at_token(filename, code, "Insufficient keyword count in function declaration.", without_whitespace[0])

    is_async = len(names_in_row) == 4 
    can_be_async = len(names_in_row) == 3 and not other_names
    if not is_async and can_be_async: # can be one of two forms: 

        return FunctionDefinition(      # func name(arg))
            keywords = names_in_row[:1],  
            name = names_in_row[1],
            args = names_in_row[2:],
            code = statements_inside_scope,
            is_async = False
        ), FunctionDefinition(           # async func name()
            keywords = names_in_row[:2],
            name = names_in_row[2],
            args = [],
            code = statements_inside_scope,
            is_async = True
        )

    elif is_async:
        
        # async func name (arg, arg, arg) ... 
        return FunctionDefinition(  
            keywords = names_in_row[:2],
            name = names_in_row[2],
            args = names_in_row[3:] + other_names,
            code = statements_inside_scope,
            is_async = True
        ),

    else:

        # func name (arg, arg, arg)
        return FunctionDefinition(
            keywords = names_in_row[:1],
            name = names_in_row[1],
            args = names_in_row[2:] + other_names,
            code = statements_inside_scope,
            is_async = False
        ),

def create_scoped_code_statement(filename: str, tokens: list[Token], without_whitespace: list[Token], code: str) -> tuple[CodeStatement, ...]:

    # this means that a scope is detected in the statement
    ends_with_punc = tokens[-1].type in {TokenType.BANG, TokenType.QUESTION}
    if tokens[-1 - int(ends_with_punc)].type != TokenType.R_CURLY:  # end will never be whitespace 
        raise_error_at_token(filename, code, "End of statement with open scope must close the scope.", tokens[-1])
    if without_whitespace[0].type != TokenType.NAME:
        raise_error_at_token(filename, code, "Scoped code statement must start with a keyword.", without_whitespace[0])

    # at this point, this can be the "when" keyword, a class dec, a function call, or an if statement
    scope_open_index = [t.type == TokenType.L_CURLY for t in tokens].index(True)
    stuff_inside_scope = tokens[scope_open_index + 1 : len(tokens) - ends_with_punc - 1]
    statements_inside_scope = generate_syntax_tree(filename, stuff_inside_scope, code)
    
    # see the function pointer -> immediately know
    can_be_function = any([t.type == TokenType.FUNC_POINT for i, t in enumerate(without_whitespace) if i < scope_open_index])

    # now finally, check for classes
    can_be_class = without_whitespace[0].type == TokenType.NAME and \
                   without_whitespace[1].type == TokenType.NAME and \
                   without_whitespace[2].type == TokenType.L_CURLY

    # finally finally, check for the after or when statement -- this will have identical syntax to the conditional so there 
    # is no point in doing anything extra special

    # this dude is seperated to another function because the same code is reused in () => ... functions (no scope)
    possibilities = []
    if can_be_function:
        return create_function_definition(filename, without_whitespace, code, statements_inside_scope)

    if can_be_class:

        # build thing for the class statement
        possibilities.append(ClassDeclaration(
            keyword = without_whitespace[0],
            name = without_whitespace[1],
            code = statements_inside_scope
        ))

    possibilities.extend([
        Conditional(
            keyword = without_whitespace[0],
            expression = tokens[int(tokens[0].type == TokenType.WHITESPACE) + 1 : scope_open_index],
            code = statements_inside_scope
        ), AfterStatement(
            keyword = without_whitespace[0],
            expression = tokens[int(tokens[0].type == TokenType.WHITESPACE) + 1 : scope_open_index],
            code = statements_inside_scope
        ), WhenStatement(
            keyword = without_whitespace[0],
            expression = tokens[int(tokens[0].type == TokenType.WHITESPACE) + 1 : scope_open_index],
            code = statements_inside_scope
        )
    ]) 
    return tuple(possibilities)

def is_proper_comma_list(without_whitespace: list[Token], accepted_tokens: frozenset[TokenType] = frozenset({TokenType.NAME})) -> bool:
    looking_for_comma = without_whitespace[0].type != TokenType.COMMA
    for t in without_whitespace[1:]:
        if not looking_for_comma and t.type not in accepted_tokens or \
           looking_for_comma and t.type != TokenType.COMMA:
            return False
        looking_for_comma = not looking_for_comma
    return True

def create_unscoped_code_statement(filename: str, tokens: list[Token], without_whitespace: list[Token], code: str) -> tuple[CodeStatement, ...]: 

    is_debug = tokens[-1].type == TokenType.QUESTION
    confidence = 0 if is_debug else len(tokens[-1].value)
    debug_level = 0 if not is_debug else len(tokens[-1].value)

    if len(l := [t for t in tokens if t.type != TokenType.WHITESPACE]) == 2:
        return (ReverseStatement(l[0], debug_level), ExpressionStatement(tokens[:-1], debug_level))

    # it's a function!!!!!!!!!!!!!!!!!
    if any(l := [t.type == TokenType.FUNC_POINT for t in tokens]):
        func_point_index = l.index(True)
        return create_function_definition(filename, without_whitespace, code, [(ReturnStatement(
            keyword = None,
            expression = tokens[func_point_index + 1 : -1],
            debug = debug_level
        ),)])

    # import statement: import name, name, name!
    can_be_import = all(t.type in {TokenType.NAME, TokenType.COMMA} for t in without_whitespace[:-1]) and \
                    len(without_whitespace) >= 3 and is_proper_comma_list(without_whitespace[1:-1]) and \
                    without_whitespace[0].type == TokenType.NAME and without_whitespace[1].type  == TokenType.NAME
   
    # export statement: export name, name, name to string/name!
    can_be_export = all(t.type in {TokenType.STRING, TokenType.NAME, TokenType.COMMA} for t in without_whitespace[:-1]) and \
                    len(without_whitespace) >= 5 and is_proper_comma_list(without_whitespace[1:-3]) and \
                    without_whitespace[0].type == TokenType.NAME and without_whitespace[1].type == TokenType.NAME and \
                    without_whitespace[-2].type in {TokenType.NAME, TokenType.STRING} and without_whitespace[-3].type == TokenType.NAME

    # let's see what can be what D:
    can_be_return = without_whitespace[0].type == TokenType.NAME
    can_be_delete = can_be_return and len(without_whitespace) == 3 and \
                    without_whitespace[1].type == TokenType.NAME and without_whitespace[2].type in {TokenType.BANG, TokenType.QUESTION}
    
    contains_equals = any(tokens_is_equal := [t.type == TokenType.EQUAL and t.value == '=' for t in tokens]) 
    can_be_var_assignment = can_be_var_declaration = contains_equals    

    # checking for single name and index for variable assignment 
    can_be_var_assignment &= len(without_whitespace) >= 4 and without_whitespace[0].type == TokenType.NAME and \
                             without_whitespace[1].type in {TokenType.EQUAL, TokenType.L_SQUARE}
    var_assignment_index: list[list[Token]] = [[]]   # is list[list] to handle multiple indexes
    bracket_layers, add_to_var_assignment_index = 0, False
    for t in tokens:

        if not can_be_var_assignment:
            break

        if t.type == TokenType.L_SQUARE:
            if bracket_layers == 0:
                add_to_var_assignment_index = True
            bracket_layers += 1
            continue
        elif t.type == TokenType.R_SQUARE:
            bracket_layers -= 1
            if bracket_layers == 0:
                add_to_var_assignment_index = False
                var_assignment_index.append([])
            continue
        elif bracket_layers == 0 and t.type == TokenType.EQUAL:  # exit when hitting the equals
            break

        if add_to_var_assignment_index:
            var_assignment_index[-1].append(t)
    var_assignment_index.pop()  # the last one will always be empty

    # checking modifiers and lifetime for varianle declaration
    names_in_row: list[Token] = []
    looking_for_lifetime, lifetime = False, None
    for t in without_whitespace:
        if not can_be_var_declaration or can_be_var_assignment:  # var assignment has a single name, therefore counteracts being decl
            break 
        if not looking_for_lifetime:
            if t.type != TokenType.NAME:
                if t.type == TokenType.LESS_THAN:
                    looking_for_lifetime = True
                    continue
                else:
                    break
            names_in_row.append(t)
        else:
            if t.type == TokenType.GREATER_THAN or not 3 <= len(names_in_row) <= 4 or not can_be_var_declaration:
                break
            if not lifetime:
                lifetime = t.value 
    
    can_be_var_declaration &= 3 <= len(names_in_row) <= 4
    
    # make a list of all possible things, starting with plain expression 
    possibilities: list[CodeStatement] = [ExpressionStatement(tokens[:-1], debug_level)] 
    if can_be_return:
        possibilities.append(ReturnStatement(
            keyword = without_whitespace[0],   # should be the same as tokens[0].value but this makes me feel safe
            expression = tokens[1:-1],
            debug = debug_level
        ))
    if can_be_delete:
        possibilities.append(DeleteStatement(
            keyword = without_whitespace[0],
            name = without_whitespace[1],
            debug = debug_level
        ))
    if can_be_import:
        possibilities.append(ImportStatement(
            keyword = without_whitespace[0],
            names = [t for t in without_whitespace[1:-1] if t.type == TokenType.NAME],
            debug = debug_level
        ))
    if can_be_export:
        possibilities.append(ExportStatement(
            export_keyword = without_whitespace[0],
            names = [t for t in without_whitespace[1:-3] if t.type == TokenType.NAME],
            to_keyword = without_whitespace[-3],
            target_file = without_whitespace[-2],
            debug = debug_level
        ))
    if can_be_var_declaration:
        possibilities.append(VariableDeclaration(
            name = names_in_row[-1],
            modifiers = names_in_row[:-1],
            lifetime = lifetime, 
            expression = tokens[tokens_is_equal.index(True) + 1 : -1],   # the end should be a puncutation
            confidence = confidence,
            debug = debug_level, 
        ))
    if can_be_var_assignment:
        possibilities.append(VariableAssignment(
            name = without_whitespace[0],
            expression = tokens[tokens_is_equal.index(True) + 1 : -1],
            debug = debug_level, 
            indexes = var_assignment_index,
            confidence = confidence
        ))
    return tuple(possibilities)

def generate_syntax_tree(filename: str, tokens: list[Token], code: str) -> list[tuple[CodeStatement, ...]]: 
    """ Split the code up into lines, which are then parsed and shit """

    assert_proper_indentation(filename, tokens, code)
    statements = split_into_statements(tokens)
    removed_hints = remove_type_hints(filename, code, statements)
    final_statements = []
    
    # now we need to perform pattern matching on each list of statements
    for tokens in removed_hints:

        without_whitespace = [t for t in tokens if t.type != TokenType.WHITESPACE]

        try:
            # contains an open scope :)
            if any(t.type == TokenType.L_CURLY for t in tokens):
                final_statements.append(create_scoped_code_statement(
                    filename, tokens, without_whitespace, code
                ))

            else:  
                final_statements.append(create_unscoped_code_statement(
                    filename, tokens, without_whitespace, code
                ))
            
            # exit if some possiblity was found 
            if final_statements[-1]:
                continue
        except IndexError:  # i have no idea what kind of errors are going to be rasied here
            pass
        raise_error_at_line(filename, code, without_whitespace[0].line, "Error parsing statement. I have no idea what went wrong, double check it and try again.")
        
    return final_statements



from __future__ import annotations
import functools

import math
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import Callable, Optional, Union
from dreamberd.base import NonFormattedError

from dreamberd.processor.syntax_tree import CodeStatement

FLOAT_TO_INT_PREC = 0.00000001
def is_int(x: Union[float, int]) -> bool:
    return min(x % 1, 1 - x % 1) < FLOAT_TO_INT_PREC

def db_not(x: DreamberdBoolean) -> DreamberdBoolean:
    if x.value is None:
        return DreamberdBoolean(None)
    return DreamberdBoolean(not x.value)

def db_list_push(self: DreamberdList, val: Value) -> None:
    self.values.append(val) 
    self.create_namespace()  # update the length

def db_list_pop(self: DreamberdList, index: Union[DreamberdNumber, DreamberdSpecialBlankValue]) -> Value:
    if isinstance(index, DreamberdSpecialBlankValue):
        retval = self.values.pop()
        self.create_namespace()
        return retval
    elif not isinstance(index, DreamberdNumber) or not is_int(index.value):
        raise NonFormattedError("Expected integer for list popping.")
    elif not -1 <= index.value <= len(self.values) - 1:
        raise NonFormattedError("Indexing out of list bounds.")
    retval = self.values.pop(round(index.value) + 1)
    self.create_namespace()
    return retval

def db_str_push(self: DreamberdString, val: Value) -> None:
    val_str = db_to_string(val).value
    self.value += val_str 
    self.create_namespace()  # update the length

def db_str_pop(self: DreamberdString, index: Union[DreamberdNumber, DreamberdSpecialBlankValue]) -> Value:
    if isinstance(index, DreamberdSpecialBlankValue):
        retval = self.value[-1]
        self.value = self.value[:-1]
        return DreamberdString(retval)
    elif not isinstance(index, DreamberdNumber) or not is_int(index.value):
        raise NonFormattedError("Expected integer for string popping.")
    elif not -1 <= index.value <= len(self.value) - 1:
        raise NonFormattedError("Indexing out of string bounds.")
    index_val = round(index.value) + 1
    retval = self.value[index_val]
    self.value = self.value[:index_val] + self.value[index_val + 1:]
    return DreamberdString(retval)

# class Value(metaclass=ABCMeta):   # TODO POTENTIALLY DO THIS TO ALLOW FOR MORE OBJECTS WITHOUT MUCH HASSLE
#     @abstractmethod 
#     def to_bool(self) -> Value: pass
#     @abstractmethod 
#     def to_num(self) -> Value: pass
#     @abstractmethod 
#     def to_str(self) -> Value: pass

class Value():  # base class for shit  
    pass

class DreamberdMutable(Value):  # mutable values
    pass

class DreamberdIndexable(Value, metaclass=ABCMeta):
    
    @abstractmethod 
    def access_index(self, index: Value) -> Value: pass

    @abstractmethod
    def assign_index(self, index: Value, val: Value) -> None: pass

class DreamberdNamespaceable(Value, metaclass=ABCMeta):
    namespace: dict[str, Union[Name, Variable]]

@dataclass 
class DreamberdFunction(Value):  
    args: list[str]
    code: list[tuple[CodeStatement, ...]]
    is_async: bool

@dataclass
class BuiltinFunction(Value):
    arg_count: int
    function: Callable
    modifies_caller: bool = False

@dataclass 
class DreamberdList(DreamberdIndexable, DreamberdNamespaceable, DreamberdMutable):
    values: list[Value]
    namespace: dict[str, Union[Name, Variable]] = field(default_factory=dict)

    def __post_init__(self):
        self.create_namespace(False)

    def create_namespace(self, is_update: bool = True) -> None:

        if not is_update:
            self.namespace = {
                'push': Name('push', BuiltinFunction(2, db_list_push, True)),
                'pop': Name('pop', BuiltinFunction(2, db_list_pop, True)),
                'length': Name('length', DreamberdNumber(len(self.values))),
            }
        elif is_update:
            self.namespace |= {
                'length': Name('length', DreamberdNumber(len(self.values))),
            }

    def access_index(self, index: Value) -> Value:
        if not isinstance(index, DreamberdNumber):
            raise NonFormattedError("Cannot index a list with a non-number value.")
        if not is_int(index.value):
            raise NonFormattedError("Expected integer for list indexing.")
        elif not -1 <= index.value <= len(self.values) - 1:
            raise NonFormattedError("Indexing out of list bounds.")
        return self.values[round(index.value) + 1]

    def assign_index(self, index: Value, val: Value) -> None:
        if not isinstance(index, DreamberdNumber):
            raise NonFormattedError("Cannot index a list with a non-number value.")
        if is_int(index.value):
            if not -1 <= index.value <= len(self.values) - 1:
                raise NonFormattedError("Indexing out of list bounds.")
            self.values[round(index.value) + 1] = val
        else:  # assign in the middle of the array
            nearest_int_down = round(max((index.value + 1) // 1, 0))
            self.values[nearest_int_down:nearest_int_down] = [val]
            self.create_namespace()

@dataclass(unsafe_hash=True)
class DreamberdNumber(DreamberdIndexable, DreamberdMutable):
    value: Union[int, float]

    def _get_self_str(self) -> str:
        return str(self.value).replace('.', '').replace('-', '')

    def access_index(self, index: Value) -> Value:
        self_val_str = self._get_self_str()
        if not isinstance(index, DreamberdNumber):
            raise NonFormattedError("Cannot index a number with a non-number value.")
        if not is_int(index.value):
            raise NonFormattedError("Expected integer for number indexing.")
        elif not -1 <= index.value <= len(self_val_str) - 1:
            raise NonFormattedError("Indexing out of number bounds.")
        return DreamberdNumber(int(self_val_str[round(index.value) + 1]))

    def assign_index(self, index: Value, val: Value) -> None:
        self_val_str = self._get_self_str()
        sign = self.value / abs(self.value)
        if not is_int(self.value):
            raise NonFormattedError("Cannot assign into a non-interger number.")
        if not isinstance(index, DreamberdNumber):
            raise NonFormattedError("Cannot index a number with a non-number value.")
        if not isinstance(val, DreamberdNumber) or not is_int(val.value) or not 0 <= val.value <= 9:
            raise NonFormattedError("Cannot assign into a number with a non-integer value.")
        if is_int(index.value):
            if not -1 <= index.value <= len(self_val_str) - 1:
                raise NonFormattedError("Indexing out of number bounds.")
            index_num = round(index.value) + 1
            self.value = sign * int(self_val_str[:index_num] + str(round(val.value)) + self_val_str[index_num + 1:])
        else:  # assign in the middle of the array
            index_num = max((index.value + 1) // 1, 0)
            self.value = sign * int(self_val_str[:index_num] + str(round(val.value)) + self_val_str[index_num:])

@dataclass(unsafe_hash=True)
class DreamberdString(DreamberdIndexable, DreamberdNamespaceable, DreamberdMutable):
    value: str = field(hash=True)
    namespace: dict[str, Union[Name, Variable]] = field(default_factory=dict, hash=False)

    def __post_init__(self):
        self.create_namespace(False)

    def create_namespace(self, is_update: bool = True):

        if not is_update:
            self.namespace |= {
                'push': Name('push', BuiltinFunction(2, db_str_push, True)),
                'length': Name('length', DreamberdNumber(len(self.value))),
            }
        else:
            self.namespace['length'] = Name('length', DreamberdNumber(len(self.value)))

    def access_index(self, index: Value) -> Value:
        if not isinstance(index, DreamberdNumber):
            raise NonFormattedError("Cannot index a string with a non-number value.")
        if not is_int(index.value):
            raise NonFormattedError("Expected integer for string indexing.")
        elif not -1 <= index.value <= len(self.value) - 1:
            raise NonFormattedError("Indexing out of string bounds.")
        return DreamberdString(self.value[round(index.value) + 1])

    def assign_index(self, index: Value, val: Value) -> None:
        if not isinstance(index, DreamberdNumber):
            raise NonFormattedError("Cannot index a string with a non-number value.")
        val_str = db_to_string(val).value
        if is_int(index.value):
            if not -1 <= index.value <= len(self.value) - 1:
                raise NonFormattedError("Indexing out of string bounds.")
            index_num = round(index.value) + 1
            self.value = self.value[:index_num] + val_str + self.value[index_num + 1:]
        else:  # assign in the middle of the array
            index_num = max((index.value + 1) // 1, 0)
            self.value = self.value[:index_num] + val_str + self.value[index_num:]
        self.create_namespace()

@dataclass 
class DreamberdBoolean(Value):
    value: Optional[bool]  # none represents maybe?

@dataclass 
class DreamberdUndefined(Value):
    pass
@dataclass 
class DreamberdSpecialBlankValue(Value):
    pass

@dataclass 
class DreamberdObject(DreamberdNamespaceable, Value):
    class_name: str
    namespace: dict[str, Union[Name, Variable]] = field(default_factory=dict)

@dataclass 
class DreamberdMap(DreamberdIndexable, Value):
    self_dict: dict[Union[int, float, str], Value]

    def access_index(self, index: Value) -> Value:
        if not isinstance(index, (DreamberdString, DreamberdNumber)):
            raise NonFormattedError("Keys of a map must be an index or a number.")
        return self.self_dict[index.value]

    def assign_index(self, index: Value, val: Value) -> None:
        if not isinstance(index, (DreamberdString, DreamberdNumber)):
            raise NonFormattedError("Keys of a map must be an index or a number.")
        self.self_dict[index.value] = val

@dataclass 
class DreamberdKeyword(Value):
    value: str

@dataclass 
class DreamberdPromise(Value):
    value: Optional[Value]

@dataclass
class Name:
    name: str
    value: Value

@dataclass 
class VariableLifetime:
    value: Value
    lines_left: int 
    confidence: int
    can_be_reset: bool
    can_edit_value: bool

@dataclass
class Variable:
    name: str 
    lifetimes: list[VariableLifetime]
    prev_values: list[Value]

    @property 
    def can_be_reset(self) -> bool:
        if self.lifetimes:
            return self.lifetimes[0].can_be_reset
        raise NonFormattedError("Variable is undefined.")

    @property 
    def can_edit_value(self) -> bool:
        if self.lifetimes:
            return self.lifetimes[0].can_edit_value
        raise NonFormattedError("Variable is undefined.")

    def add_lifetime(self, value: Value, confidence: int, duration: int, can_be_reset: bool, can_edit_value: bool) -> None:
        for i in range(len(self.lifetimes) + 1):
            if i == len(self.lifetimes) or self.lifetimes[i].confidence == confidence:
                if i == 0:
                    self.prev_values.append(self.value)
                self.lifetimes[i:i] = [VariableLifetime(value, duration, confidence, can_be_reset, can_edit_value)]
                break

    def clear_outdated_lifetimes(self) -> None:
        remove_indeces = []
        for i, l in enumerate(self.lifetimes):
            if l.lines_left == 0:
                remove_indeces.append(i)
        for i in reversed(remove_indeces):
            del self.lifetimes[i]

    @property
    def value(self) -> Value:
        if self.lifetimes:
            return self.lifetimes[0].value
        raise NonFormattedError("Variable is undefined.")
    
def all_function_keywords() -> list[str]:

    # this code boutta be crazy
    # i refuse to use the builtin combinations
    keywords = set()
    for f in range(2):
        for u in range(2):
            for n in range(2):
                for c in range(2):
                    for t in range(2):
                        for i in range(2):
                            for o in range(2):
                                for n_ in range(2):
                                    keywords.add("".join([c * i for c, i in zip('function', [f,u,n,c,t,i,o,n_])]) or 'fn')    # the `or` allows the empty string to not count
    return list(keywords)

FUNCTION_KEYWORDS = all_function_keywords()
KEYWORDS = {kw: Name(kw, DreamberdKeyword(kw)) for kw in 
    ['class', 'className', 'after', 'const', 'var', 'when', 'if', 'async', 
     'return', 'delete', 'await', 'previous', 'next', 'reverse', 'export', 'import'] + FUNCTION_KEYWORDS}

############################################
##           DREAMBERD BUILTINS           ##
############################################

# this is for functions that return the same value, like current or new
def db_identity(val: Value) -> Value:
    return val

def db_map() -> DreamberdMap:
    return DreamberdMap({})

def db_to_boolean(val: Value) -> DreamberdBoolean:
    return_bool = None
    match val: 
        case DreamberdString():
            return_bool = bool(val.value.strip()) or (None if len(val.value) else False)
        case DreamberdNumber():  # maybe if it is 0.xxx, false if it is 0, true if anything else
            return_bool = bool(round(val.value)) or (None if abs(val.value) > FLOAT_TO_INT_PREC else False)
        case DreamberdList():
            return_bool = bool(val.values)
        case DreamberdMap():
            return_bool = bool(val.self_dict)
        case DreamberdBoolean():
            return_bool = val.value
        case DreamberdUndefined():
            return_bool = False
        case DreamberdFunction() | DreamberdObject() | DreamberdKeyword():
            return_bool = None  # maybe for these cause im mischevious
    return DreamberdBoolean(return_bool)

def db_to_string(val: Value) -> DreamberdString:
    return_string = str(val)
    match val:
        case DreamberdString():
            return_string = val.value
        case DreamberdList():
            return_string = f"[{', '.join([db_to_string(v).value for v in val.values])}]"
        case DreamberdBoolean():
            return_string = "true"  if val.value else \
                            "maybe" if val.value is None else "false"
        case DreamberdNumber():
            return_string = str(val.value)
        case DreamberdFunction(): 
            return_string = f"<function ({', '.join(val.args)})>"
        case DreamberdObject():
            return_string = f"<object {val.class_name}>" 
        case DreamberdUndefined():
            return_string = "undefined"
        case DreamberdKeyword():
            return_string = val.value
    return DreamberdString(return_string)

def db_print(*vals: Value) -> None:
    print(*[db_to_string(v).value for v in vals])

def db_to_number(val: Value) -> DreamberdNumber:
    return_number = 0
    match val:
        case DreamberdNumber():
            return_number = val.value
        case DreamberdString():
            return_number = float(val.value)
        case DreamberdUndefined():
            return_number = 0 
        case DreamberdBoolean():
            return_number = int(val.value is not None and val.value) + (val.value is None) * 0.5
        case DreamberdList():
            if val.values:
                raise NonFormattedError("Cannot turn a non-empty list into a number.")
            return_number = 0 
        case DreamberdMap():
            if val.self_dict:
                raise NonFormattedError("Cannot turn a non-empty map into a number.")
            return_number = 0 
        case _:
            raise NonFormattedError(f"Cannot turn type {type(val).__name__} into a number.")
    return DreamberdNumber(return_number)

def db_signal(starting_value: Value) -> Value:
    obj = Name('', starting_value)
    def signal_func(setter_val: Value) -> Optional[Value]:
        nonlocal obj
        if isinstance(setter_val, DreamberdSpecialBlankValue):
            return obj.value
        obj.value = setter_val
    return BuiltinFunction(1, signal_func)

def db_exit() -> None:
    exit()

def __math_function_decorator(func: Callable):
    @functools.wraps(func)
    def inner(*args) -> DreamberdNumber:  # no kwargs
        for arg in args:
            if not isinstance(arg, DreamberdNumber):
                raise NonFormattedError("Cannot pass in a non-number value into a math function.")
        return DreamberdNumber(func(*[arg.value for arg in args]))
    return inner

def __number_function_maker(num: int) -> BuiltinFunction:
    def the_func(n: DreamberdNumber) -> DreamberdNumber:
        nonlocal num
        if not isinstance(n, DreamberdNumber):
            raise NonFormattedError(f"Expected a number in the ones digit. Instead received a {type(n).__name__}")
        return DreamberdNumber(num + n.value)
    return BuiltinFunction(1, the_func)

# get ready, this is boutta be crazy
MATH_FUNCTION_KEYWORDS = {
    name: Name(name, BuiltinFunction((-1 if any([arg[0] == '*' and len(arg) > 1 for arg in v.__text_signature__[1:-1].split(', ')]) 
                                         else len([arg for arg in v.__text_signature__[1:-1].split(', ') if arg.isalpha()])) 
        if v.__text_signature__ else 1 if name == 'log' else -1, __math_function_decorator(v)) 
    if isinstance(v := getattr(math, name), type(math.ulp)) else DreamberdNumber(v)) for name in dir(math) if not name.startswith('__')
}  # the frick is this
BUILTIN_FUNCTION_KEYWORDS = {
    "new": Name("new", BuiltinFunction(1, db_identity)),
    "current": Name("current", BuiltinFunction(1, db_identity)),
    "Map": Name("Map", BuiltinFunction(0, db_map)),
    "Boolean": Name("Boolean", BuiltinFunction(1, db_to_boolean)),
    "String": Name("String", BuiltinFunction(1, db_to_string)),
    "print": Name("print", BuiltinFunction(-1, db_print)),
    "exit": Name("exit", BuiltinFunction(0, db_exit)),
    "Number": Name("Number", BuiltinFunction(1, db_to_number)),
    "use": Name("use", BuiltinFunction(1, db_signal))
}
BUILTIN_VALUE_KEYWORDS = {
    "true": Name("true", DreamberdBoolean(True)),
    "maybe": Name("maybe", DreamberdBoolean(None)),
    "false": Name("false", DreamberdBoolean(False)),
    "undefined": Name("undefined", DreamberdUndefined()),
    "": Name("", DreamberdSpecialBlankValue())
}
NUMBER_NAME_KEYWORDS = {
    name: Name(name, DreamberdNumber(num)) for num, name in enumerate(["zero", "one", "two", "three", "four", "five", "six", "seven", "eight", "nine", "ten",
     "eleven", "twelve", "thirteen", "fourteen", "fifteen", "sixteen", "seventeen", "eighteen", "ninteen"])
} | {
    name: Name(name, __number_function_maker(num)) for num, name in zip(range(20, 100, 10), ["twenty", "thirty", "forty", "fifty", "sixty", "seventy", "eighty", "ninety"])
}  # this is so cursed

KEYWORDS |= BUILTIN_FUNCTION_KEYWORDS | BUILTIN_VALUE_KEYWORDS | MATH_FUNCTION_KEYWORDS |\
            NUMBER_NAME_KEYWORDS
